
from enum import Enum as Enumerate

from sqlalchemy import Column, Text, String, Enum
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound
from vantage6.server.model.base import Base, DatabaseSessionManager


class Operation(Enumerate):
    VIEW = "v"
    EDIT = "e"
    CREATE = "c"
    DELETE = "d"


class Scope(Enumerate):
    OWN = "own"
    ORGANIZATION = "org"
    COLLABORATION = "col"
    GLOBAL = "glo"


class Rule(Base):
    """Rules to determine permissions in an API endpoint.
    """

    # fields
    name = Column(Text)
    operation = Column(Enum(Operation))
    scope = Column(Enum(Scope))
    description = Column(String)

    # relationships
    roles = relationship("Role", back_populates="rules",
                         secondary="role_rule_association")
    users = relationship("User", back_populates="rules",
                         secondary="UserPermission")

    @classmethod
    def get_by_(cls, name: str, scope: str, operation: str):
        session = DatabaseSessionManager.get_session()
        try:
            return session.query(cls).filter_by(
                name=name,
                operation=operation,
                scope=scope
            ).first()
        except NoResultFound:
            return None

    def __repr__(self):
        return (
            f"<Rule "
            f"{self.id}: '{self.name}', "
            f"operation: {self.operation}, "
            f"scope: {self.scope}"
            ">"
        )



from sqlalchemy import Column, Integer, ForeignKey, Table

from .base import Base


Member = Table(
    'Member',
    Base.metadata,
    Column('organization_id', Integer, ForeignKey('organization.id')),
    Column('collaboration_id', Integer, ForeignKey('collaboration.id'))
)



from vantage6.server.model.rule import Operation, Scope


# TODO BvB 22-06-07: we now have to pass this 'db' module as argument to a
# function because that module has a connection to the database. This should
# not be necessary. Fix that after fixing the circular imports described in
# https://github.com/vantage6/vantage6/issues/53
def get_default_roles(db):
    # Define default roles
    # 1. Root user
    SUPER_ROLE = {
        'name': 'Root',
        'description': "Super role",
        'rules': db.Rule.get()
    }
    # 2. Role for viewing organization resources
    VIEWER_RULES = [
        db.Rule.get_by_('user', Scope.OWN, Operation.EDIT),
        db.Rule.get_by_('user', Scope.OWN, Operation.DELETE),
        db.Rule.get_by_('user', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('organization', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('organization', Scope.COLLABORATION, Operation.VIEW),
        db.Rule.get_by_('collaboration', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('role', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('node', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('task', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('result', Scope.ORGANIZATION, Operation.VIEW),
        db.Rule.get_by_('port', Scope.ORGANIZATION, Operation.VIEW),
    ]
    VIEWER_ROLE = {
        'name': 'Viewer',
        'description': "Can manage their own account and view resources "
                       "related to their organization",
        'rules': VIEWER_RULES
    }
    # 3. Researcher role
    RESEARCHER_RULES = VIEWER_RULES + [
        db.Rule.get_by_('task', Scope.ORGANIZATION, Operation.CREATE),
        db.Rule.get_by_('task', Scope.ORGANIZATION, Operation.DELETE),
    ]
    RESEARCHER_ROLE = {
        'name': 'Researcher',
        'description': "Can perform tasks, manage their own account, and "
                       "view resources related to their organization",
        'rules': RESEARCHER_RULES
    }
    # 4. Organization administrator role
    ORG_ADMIN_RULES = RESEARCHER_RULES + [
        db.Rule.get_by_('user', Scope.ORGANIZATION, Operation.CREATE),
        db.Rule.get_by_('user', Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_('user', Scope.ORGANIZATION, Operation.DELETE),
        db.Rule.get_by_('organization', Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_('role', Scope.ORGANIZATION, Operation.CREATE),
        db.Rule.get_by_('role', Scope.ORGANIZATION, Operation.EDIT),
        db.Rule.get_by_('role', Scope.ORGANIZATION, Operation.DELETE),
        db.Rule.get_by_('node', Scope.ORGANIZATION, Operation.EDIT),
    ]
    ORG_ADMIN_ROLE = {
        'name': 'Organization Admin',
        'description':
            "Can manage an organization including its users, roles, and nodes."
            " Also has all permissions of a researcher.",
        'rules': ORG_ADMIN_RULES
    }
    # 4. Collaboration administrator role
    COLLAB_ADMIN_RULES = ORG_ADMIN_RULES + [
        db.Rule.get_by_('user', Scope.GLOBAL, Operation.VIEW),
        db.Rule.get_by_('user', Scope.GLOBAL, Operation.CREATE),
        db.Rule.get_by_('user', Scope.GLOBAL, Operation.EDIT),
        db.Rule.get_by_('organization', Scope.GLOBAL, Operation.VIEW),
        db.Rule.get_by_('organization', Scope.GLOBAL, Operation.EDIT),
        db.Rule.get_by_('collaboration', Scope.GLOBAL, Operation.VIEW),
        db.Rule.get_by_('collaboration', Scope.GLOBAL, Operation.EDIT),
        db.Rule.get_by_('role', Scope.GLOBAL, Operation.VIEW),
        db.Rule.get_by_('node', Scope.GLOBAL, Operation.CREATE),
        db.Rule.get_by_('node', Scope.GLOBAL, Operation.VIEW),
        db.Rule.get_by_('node', Scope.GLOBAL, Operation.DELETE),
    ]
    COLLAB_ADMIN_ROLE = {
        'name': 'Collaboration Admin',
        'description':
            "Can manage an collaboration including its organization and users."
            " Also has permissions of an organization admin.",
        'rules': COLLAB_ADMIN_RULES
    }
    # Combine all in array
    return [
        SUPER_ROLE, VIEWER_ROLE, RESEARCHER_ROLE, ORG_ADMIN_ROLE,
        COLLAB_ADMIN_ROLE
    ]



# -*- coding: utf-8 -*-
import logging
import os
import json
import base64
import hashlib
import re
import urllib.parse as urlparse

from http import HTTPStatus
from flask.globals import g
from flask import request
import requests
from requests_oauthlib import OAuth2Session

from vantage6.common import logger_name
from vantage6.server.resource import with_node, ServicesResources


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api, api_base, services):
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        VPNConfig,
        path,
        endpoint='vpn_config',
        methods=('GET',),
        resource_class_kwargs=services
    )

    api.add_resource(
        VPNConfig,
        path + '/update',
        endpoint='vpn_config_update',
        methods=('POST',),
        resource_class_kwargs=services
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class VPNConfig(ServicesResources):

    @with_node
    def get(self):
        """Get OVPN configuration file
        ---
        description: >-
          Returns the contents of an OVPN configuration file. This
          configuration allows the node to connect to the VPN server.\n

          This endpoint is not accessible for users, but only for
          authenticated nodes.

        responses:
          200:
            description: Ok
          501:
            description: This server has no VPN service
          503:
            description: The VPN server cannot be reached

        security:
            - bearerAuth: []

        tags: ["VPN"]

        """
        # check if the VPN server is configured
        if not self._is_server_configured():
            return {'msg': 'This server does not support VPN'}, \
                HTTPStatus.NOT_IMPLEMENTED

        # obtain VPN config by calling EduVPN API
        vpn_connector = EduVPNConnector(self.config['vpn_server'])

        try:
            ovpn_config = vpn_connector.get_ovpn_config()
        except requests.ConnectionError as e:
            log.critical(f'Node <{g.node.id}> tries to obtain a vpn config. '
                         'However the VPN server is unreachable!')
            log.debug(e)
            return {'msg': 'VPN server unreachable'}, \
                HTTPStatus.SERVICE_UNAVAILABLE

        return {'ovpn_config': ovpn_config}, HTTPStatus.OK

    @with_node
    def post(self):
        """ Update an OVPN configuration file
        ---
        description: >-
          Returns an OVPN configuration file with renewed keypair. This
          allows the node to connect to the VPN server again if the keypair was
          invalidated.\n

          This endpoint is not accessible for users, but only for
          authenticated nodes.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  vpn_config:
                    type: string
                    description: Current VPN config file contents with expired
                      keypair

        responses:
          200:
            description: Ok
          400:
            description: No VPN configuration found in request body
          501:
            description: This server has no VPN service
          503:
            description: The VPN server cannot be reached

        security:
            - bearerAuth: []

        tags: ["VPN"]

        """
        # check if the VPN server is configured
        if not self._is_server_configured():
            return {'msg': 'This server does not support VPN'}, \
                HTTPStatus.NOT_IMPLEMENTED

        # retrieve user based on email or username
        body = request.get_json()
        vpn_config = body.get("vpn_config")
        if not vpn_config:
            return {"msg": "vpn_config is missing!"}, \
                HTTPStatus.BAD_REQUEST

        # obtain VPN config by calling EduVPN API
        vpn_connector = EduVPNConnector(self.config['vpn_server'])

        try:
            ovpn_config = vpn_connector.refresh_keypair(vpn_config)
        except requests.ConnectionError as e:
            log.critical(f'Node <{g.node.id}> tries to obtain a vpn config. '
                         'However the VPN server is unreachable!')
            log.debug(e)
            return {'msg': 'VPN server unreachable'}, \
                HTTPStatus.SERVICE_UNAVAILABLE

        return {'ovpn_config': ovpn_config}, HTTPStatus.OK

    def _is_server_configured(self) -> bool:
        """ Check if vpn server is available in configuration"""
        if 'vpn_server' not in self.config:
            log.debug(f'Node <{g.node.id}> tries to obtain a vpn config but '
                      'this server has not configured a VPN server!')
            return False
        return True


class EduVPNConnector:

    def __init__(self, vpn_config):
        """
        Provides API access to the VPN server

        Parameters
        ----------
        vpn_config : dict
            Contains the VPN server info
        """
        self.config = vpn_config
        self.session = OAuth2Session(vpn_config['client_id'])

        self.PORTAL_URL = self.config['url'][:-1] \
            if self.config['url'].endswith('/') \
            else self.config['url']
        self.API_URL = f'{self.PORTAL_URL}/api.php'

    def get_ovpn_config(self) -> str:
        """
        Obtain a configuration file from the VPN server. This file contains
        all details needed for a node to connect to the VPN server.

        Returns
        -------
        str (ovpn format)
            Open-vpn configuration file
        """

        # obtain access token if not set
        self.set_access_token()

        # get the user's profile id
        log.debug("Getting EduVPN profile information")
        profile = self.get_profile()
        profile_id = profile['profile_list']['data'][0]['profile_id']

        # get the OVPN configuration for the selected user profile
        log.debug("Obtaining OpenVPN configuration template")
        ovpn_config = self.get_config(profile_id)

        # add the keypair
        return self._add_key_pair(ovpn_config=ovpn_config)

    def refresh_keypair(self, ovpn_config: str) -> str:
        """
        Obtain a new keypair from the VPN server and refreshes the keypair so
        that the configuration file can be used to connect the VPN server
        again.

        Returns
        -------
        str (ovpn format)
            Open-vpn configuration file
        """
        # obtain access token if not set
        self.set_access_token()

        # remove current keypair from the file
        ovpn_config = self._remove_keypair(ovpn_config)

        # add the keypair
        return self._add_key_pair(ovpn_config=ovpn_config)

    def _add_key_pair(self, ovpn_config: str) -> str:
        """
        Obtain a keypair from the VPN server and add it to the configuration

        Returns
        -------
        str (ovpn format)
            Open-vpn configuration file
        """
        # get a key and certificate for the client
        log.debug("Obtaining OpenVPN key-pair")
        cert, key = self.get_key_pair()

        # add the client credentials to the ovpn file
        log.debug('Parsing VPN configuration file')
        ovpn_config = self._insert_keypair_into_config(ovpn_config, cert, key)
        return ovpn_config

    def set_access_token(self):
        """ Obtain an access token to enable access to EduVPN API """
        if self.session.token:
            log.debug("Acquiring EduVPN access token")
            return
        # set PKCE data (code challenge and code verifier)
        log.debug("Setting PKCE challenge")
        self._set_pkce()
        # login to the EduVPN portal
        log.debug("Logging in to EduVPN portal")
        self._login()
        # call the authorization route of EduVPN to get authorization code
        log.debug("Authorizing to EduVPN portal")
        self._authorize()
        # use authorization code to obtain token
        log.debug("Obtaining token from EduVPN portal")
        self.session.token = self._get_token()

    def _set_pkce(self):
        """ Generate PKCE code verifier and challenge """
        # set PKCE verifier
        self.code_verifier = \
            base64.urlsafe_b64encode(os.urandom(40)).decode('utf-8')
        self.code_verifier = re.sub('[^a-zA-Z0-9]+', '', self.code_verifier)

        # set PKCE challenge based on the verifier
        self.code_challenge = \
            hashlib.sha256(self.code_verifier.encode('utf-8')).digest()
        self.code_challenge = \
            base64.urlsafe_b64encode(self.code_challenge).decode('utf-8')
        self.code_challenge = self.code_challenge.replace('=', '')

    def _login(self):
        """ Login to the EduVPN user portal in the requests session """
        post_data = {
            'userName': self.config['portal_username'],
            'userPass': self.config['portal_userpass'],
            '_form_auth_redirect_to': self.config['url']
        }
        self.session.post(
            f'{self.PORTAL_URL}/_form/auth/verify', data=post_data
        )

    def _authorize(self):
        """ Call authorization route of EduVPN to get authorization code """
        params = {
            'client_id': self.config['client_id'],
            'redirect_uri': self.config['redirect_url'],
            'response_type': 'code',
            'scope': 'config',
            'state': 'some_string',
            'code_challenge_method': 'S256',
            'code_challenge': self.code_challenge,
        }

        # Call authorization route, but prevent redirect. By preventing the
        # redirect, the authorization code can be used in the current session
        # which allows us to use the PKCE code verifier without having to save
        # it in e.g. database or session state.
        response = self.session.get(
            f'{self.PORTAL_URL}/_oauth/authorize',
            params=params,
            allow_redirects=False
        )

        # get the authorization token from the request headers
        redirected_uri = response.headers['Location']
        parsed_url = urlparse.urlparse(redirected_uri)
        self.code = urlparse.parse_qs(parsed_url.query)['code']

    def _get_token(self):
        """ Use authorization code to obtain a token from the EduVPN portal """
        data = {
            'code': self.code,
            'grant_type': 'authorization_code',
            'redirect_uri': self.config['redirect_url'],
            'client_id': self.config['client_id'],
            'client_secret': self.config['client_secret'],
            'code_verifier': self.code_verifier,
        }

        r = self.session.post(
            f'{self.PORTAL_URL}/oauth.php/token',
            data=data,
            auth=(self.config['client_id'], self.config['client_secret'])
        )
        return json.loads(r.content.decode('utf-8'))

    def _insert_keypair_into_config(self, ovpn_config: str, cert: str,
                                    key: str):
        """
        Insert the client's key pair into the correct place into the OVPN file
        (i.e. before the <tls-crypt> field)

        Parameters
        ----------
        config : str
            The OVPN configuration information without client keys
        cert : str
            The client's certificate
        key: str
            The client's private key

        Returns
        -------
        str
            A complete OVPN config file contents, including client's key-pair
        """
        insert_loc = ovpn_config.find('<tls-crypt>')
        return (
            ovpn_config[:insert_loc] +
            '<cert>\n' + cert + '\n</cert>\n' +
            '<key>\n' + key + '\n</key>\n' +
            ovpn_config[insert_loc:]
        )

    def _remove_keypair(self, ovpn_config: str) -> str:
        """
        Remove the keypair from the configuration

        Returns
        -------
        str (ovpn format)
            Open-vpn configuration file without key pair
        """
        # keypair starts at '<cert>' and ends at '</key>\n'
        start_remove_pos = ovpn_config.find('<cert>')
        end_key = '</key>\n'
        end_remove_pos = ovpn_config.find(end_key)
        return (
            ovpn_config[0:start_remove_pos] +
            ovpn_config[end_remove_pos+len(end_key):]
        )

    def get_profile(self):
        """ Call the profile_list route of EduVPN API """
        response = self.session.get(f'{self.API_URL}/profile_list')
        return json.loads(response.content.decode('utf-8'))

    def get_config(self, profile_id):
        """ Call the profile_config route of EduVPN API

        Parameters
        ----------
        profile_id: str
            An EduVPN user's profile_id obtained from the /profile_list route
        """
        params = {
            'profile_id': profile_id
        }
        response_config = self.session.get(
            f'{self.API_URL}/profile_config', params=params
        )
        return response_config.content.decode('utf-8')

    def get_key_pair(self):
        """ Call the create_keypair route of EduVPN API """
        response_keypair = self.session.post(f'{self.API_URL}/create_keypair')
        ovpn_keypair = json.loads(response_keypair.content.decode('utf-8'))
        cert = ovpn_keypair['create_keypair']['data']['certificate']
        key = ovpn_keypair['create_keypair']['data']['private_key']
        return cert, key



# -*- coding: utf-8 -*-
import datetime
from http import HTTPStatus
import logging

from functools import wraps

from flask import g, request
from flask_restful import Resource
from flask_jwt_extended import (
    get_jwt, get_jwt_identity, jwt_required
)

from vantage6.common import logger_name
from vantage6.server import db

log = logging.getLogger(logger_name(__name__))


class ServicesResources(Resource):
    """Flask resource base class.

        Adds functionality like mail, socket, permissions and the api itself.
        Also adds common helper functions.
    """

    def __init__(self, socketio, mail, api, permissions, config):
        self.socketio = socketio
        self.mail = mail
        self.api = api
        self.permissions = permissions
        self.config = config

    @staticmethod
    def is_included(field):
        """Check that a `field` is included in the request argument context."""
        return field in request.args.getlist('include')

    def dump(self, page, schema):
        """Dump based on the request context (to paginate or not)"""
        if self.is_included('metadata'):
            return schema.meta_dump(page)
        else:
            return schema.default_dump(page)

    def response(self, page, schema):
        """Prepare a valid HTTP OK response from a page object"""
        return self.dump(page, schema), HTTPStatus.OK, page.headers

    @staticmethod
    def obtain_auth():
        """Obtain a authenticable object or dict in the case of a container."""
        if g.user:
            return g.user
        if g.node:
            return g.node
        if g.container:
            return g.container

    @staticmethod
    def obtain_organization_id():
        """Obtain the organization id from the auth that is logged in."""
        if g.user:
            return g.user.organization.id
        elif g.node:
            return g.node.organization.id
        else:
            return g.container["organization_id"]

    @classmethod
    def obtain_auth_organization(cls):
        """Obtain the organization model from the auth that is logged in."""
        return db.Organization.get(cls.obtain_organization_id())


# ------------------------------------------------------------------------------
# Helper functions/decoraters ...
# ------------------------------------------------------------------------------
def only_for(types=['user', 'node', 'container']):
    """JWT endpoint protection decorator"""
    def protection_decorator(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):

            # decode JWT-token
            identity = get_jwt_identity()
            claims = get_jwt()

            # check that identity has access to endpoint
            g.type = claims["client_type"]
            # log.debug(f"Endpoint accessed as {g.type}")

            if g.type not in types:
                msg = f"{g.type}s are not allowed to access {request.url} " \
                      f"({request.method})"
                log.warning(msg)
                raise Exception(msg)

            # do some specific stuff per identity
            g.user = g.container = g.node = None

            if g.type == 'user':
                user = get_and_update_authenticatable_info(identity)
                g.user = user
                assert g.user.type == g.type

            elif g.type == 'node':
                node = get_and_update_authenticatable_info(identity)
                g.node = node
                assert g.node.type == g.type

            elif g.type == 'container':
                g.container = identity

            else:
                raise Exception(f"Unknown entity: {g.type}")

            return fn(*args, **kwargs)
        return jwt_required()(decorator)
    return protection_decorator


def get_and_update_authenticatable_info(auth_id):
    """Get DB entity from ID and update info."""
    auth = db.Authenticatable.get(auth_id)
    auth.last_seen = datetime.datetime.utcnow()
    auth.save()
    return auth


# create alias decorators
with_user_or_node = only_for(["user", "node"])
with_user = only_for(["user"])
with_node = only_for(["node"])
with_container = only_for(["container"])


def parse_datetime(dt=None, default=None):
    if dt:
        return datetime.datetime.strptime(dt, '%Y-%m-%dT%H:%M:%S.%f')
    return default



from sqlalchemy import Column, String, Boolean, exists
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound

from vantage6.server.model.base import Base, DatabaseSessionManager


class Collaboration(Base):
    """Combination of 2 or more Organizations."""

    # fields
    name = Column(String, unique=True)
    encrypted = Column(Boolean, default=1)

    # relationships
    organizations = relationship("Organization", secondary="Member",
                                 back_populates='collaborations')
    nodes = relationship("Node", back_populates="collaboration")
    tasks = relationship("Task", back_populates="collaboration")

    def get_organization_ids(self):
        return [organization.id for organization in self.organizations]

    def get_task_ids(self):
        return [task.id for task in self.tasks]

    def get_nodes_from_organizations(self, ids):
        """Returns a subset of nodes"""
        return [n for n in self.nodes if n.organization.id in ids]

    def get_node_from_organization(self, organization):
        for node in self.nodes:
            if node.organization.id == organization.id:
                return node
        return None

    @classmethod
    def find_by_name(cls, name):
        """Find collaboration by its name.

        If multiple collaborations share the same name, the first
        collaboration found is returned."""
        session = DatabaseSessionManager.get_session()
        try:
            return session.query(cls).filter_by(name=name).first()
        except NoResultFound:
            return None

    @classmethod
    def name_exists(cls, name):
        session = DatabaseSessionManager.get_session()
        return session.query(exists().where(cls.name == name)).scalar()

    def __repr__(self):
        number_of_organizations = len(self.organizations)
        number_of_tasks = len(self.tasks)
        return (
            "<Collaboration "
            f"{self.id}: '{self.name}', "
            f"{number_of_organizations} organization(s), "
            f"{number_of_tasks} task(s)"
            ">"
        )



# -*- coding: utf-8 -*-
import logging
import datetime

import enum
import json
import sqlalchemy as sql

# TODO this file is akward...
from vantage6.server.model import (
    Base,
    Task,
    Result,
    Organization,
    User,
    Node,
    Authenticatable,
    Collaboration,
    Member,
    Permission,
    Role,
    Rule,
    UserPermission,
    role_rule_association,
    AlgorithmPort
)
from vantage6.common import logger_name
from vantage6.common.globals import STRING_ENCODING


module_name = logger_name(__name__)
log = logging.getLogger(module_name)

# DB connection session. This is used by the iPython shell (from db import
# session). It is important to mention that the flask requests obtain their
# session from `g.session` which is initialized on `pre_request`.
session = None


def jsonable(value):
    """Convert a (list of) SQLAlchemy instance(s) to native Python objects."""
    if isinstance(value, list):
        return [jsonable(i) for i in value]

    elif isinstance(value, Base):
        log.debug(f"preparing={value}")
        retval = dict()
        mapper = sql.inspect(value.__class__)

        columns = [c.key for c in mapper.columns
                   if c.key not in value._hidden_attributes]

        for column in columns:
            # log.debug(f"processing column={column}")
            column_value = getattr(value, column)

            if isinstance(column_value, enum.Enum):
                column_value = column_value.value
            elif isinstance(column_value, datetime.datetime):
                column_value = column_value.isoformat()
            elif isinstance(column_value, bytes):
                log.debug(f"decoding bytes!")
                column_value = column_value.decode(STRING_ENCODING)

            retval[column] = column_value

        return retval

    # FIXME: does it make sense to raise an exception or should base types
    #        (or other JSON-serializable types) just be returned as-is?
    raise Exception('value should be instance of db.Base or list!')


def jsonify(value):
    """Convert a (list of) SQLAlchemy instance(s) to a JSON (string)."""
    return json.dumps(jsonable(value))



from sqlalchemy import Column, String, ForeignKey, Integer, sql
from sqlalchemy.orm import relationship
from sqlalchemy.ext.hybrid import hybrid_property

from vantage6.server.model.node import Node
from vantage6.server.model.base import Base, DatabaseSessionManager


class Task(Base):
    """Central definition of a single task.

    A task can assigned in the Result for multiple organizations. The input
    of the task is different for each organization (due to the encryption).
    Therefore the input for the task is encrypted for each organization
    seperately. The task originates from an organization to which the results
    need to be encrypted, therefore the originating organization is also logged
    """

    # fields
    name = Column(String)
    description = Column(String)
    image = Column(String)
    collaboration_id = Column(Integer, ForeignKey("collaboration.id"))
    run_id = Column(Integer)
    parent_id = Column(Integer, ForeignKey("task.id"))
    database = Column(String)
    initiator_id = Column(Integer, ForeignKey("organization.id"))

    # relationships
    collaboration = relationship("Collaboration", back_populates="tasks")
    parent = relationship("Task", remote_side="Task.id", backref="children")
    results = relationship("Result", back_populates="task")
    initiator = relationship("Organization", back_populates="created_tasks")

    @hybrid_property
    def complete(self):
        return all([r.complete for r in self.results])

    def results_for_node(self, node):
        assert isinstance(node, Node), "Should be a node..."
        return [result for result in self.results if
                self.collaboration == node.collaboration and
                self.organization == node.organization]

    @classmethod
    def next_run_id(cls):
        session = DatabaseSessionManager.get_session()
        max_run_id = session.query(sql.func.max(cls.run_id)).scalar()
        if max_run_id:
            return max_run_id + 1
        else:
            return 1

    def __repr__(self):
        return (
            f"<Task "
            f"{self.id}: '{self.name}', "
            f"collaboration:{self.collaboration.name}"
            ">"
        )



import logging
import os
import inspect as class_inspect
from flask.globals import g

from sqlalchemy import Column, Integer, inspect
from sqlalchemy.orm.session import Session
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy import create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from vantage6.common import logger_name, Singleton
from vantage6.server import db


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class Database(metaclass=Singleton):
    """A singleton we can destroy, a module we cannot.

        Thats why we want a singlton. This is especially usefull when creating
        unit test in which we want fresh databases every now and then.
    """

    def __init__(self):
        self.engine = None
        self.Session = None
        self.object_session = None
        self.allow_drop_all = False

    def drop_all(self):
        if self.allow_drop_all:
            Base.metadata.drop_all(bind=self.engine)
            # Base.metadata.create_all(bind=self.engine)
            # self.Session.close()
        else:
            log.error("Cannot drop tables, configuration does not allow this!")

    def clear_data(self):
        meta = Base.metadata
        session = DatabaseSessionManager.get_session()
        for table in reversed(meta.sorted_tables):
            session.execute(table.delete())
        session.commit()
        DatabaseSessionManager.clear_session()

    def close(self):
        self.drop_all()
        self.engine = None
        self.Session = None
        self.object_session = None
        self.allow_drop_all = False
        self.URI = None

    def connect(self, uri='sqlite:////tmp/test.db', allow_drop_all=False):

        self.allow_drop_all = allow_drop_all
        self.URI = uri

        URL = make_url(uri)
        log.info("Initializing the database")
        log.debug("  driver:   {}".format(URL.drivername))
        log.debug("  host:     {}".format(URL.host))
        log.debug("  port:     {}".format(URL.port))
        log.debug("  database: {}".format(URL.database))
        log.debug("  username: {}".format(URL.username))

        # Make sure that the director for the file database exists.
        if URL.host is None and URL.database:
            os.makedirs(os.path.dirname(URL.database), exist_ok=True)

        self.engine = create_engine(uri, convert_unicode=True,
                                    pool_pre_ping=True)

        # we can call Session() to create a session, if a session already
        # exists it will return the same session (!). implicit access to the
        # Session (without calling it first). The scoped session is scoped to
        # the local thread the process is running in.
        self.session_a = scoped_session(sessionmaker(autocommit=False,
                                                     autoflush=False))
        self.session_a.configure(bind=self.engine)

        # because the Session factory returns the same session (if one exists
        # already) we need a second factory to create an alternative session.
        # this is required if we use both the flask session and the iPython.
        # Because the flask session is managed by the hooks `pre_request` and
        # `post request`. If we would use the same session for other tasks, the
        # session can be terminated unexpectedly.
        self.session_b = scoped_session(sessionmaker(autocommit=False,
                                                     autoflush=False))
        self.session_b.configure(bind=self.engine)

        # short hand to obtain a object-session.
        self.object_session = Session.object_session

        Base.metadata.create_all(bind=self.engine)
        log.info("Database initialized!")


class DatabaseSessionManager:
    """Class to manage DB sessions from.

    There are 2 different ways a session can be obtained. Either a session used
    within a request or a session used else where (e.g. iPython or within the
    application itself).

    In case of the flask-request the session is stored in the flask global `g`.
    So that it can be accessed in every endpoint.

    In all other cases the session is attached to the db module.
    """

    @staticmethod
    def in_flask_request():
        return True if g else False

    @staticmethod
    def get_session():
        if DatabaseSessionManager.in_flask_request():

            # needed for SocketIO requests
            if 'session' not in g:
                DatabaseSessionManager.new_session()

            return g.session
        else:
            # log.critical('Obtaining non flask session')
            if not db.session:
                DatabaseSessionManager.new_session()
                # log.critical('WE NEED TO MAKE A NEW ONE')

            # print(f'db.session {db.session}')
            return db.session

    @staticmethod
    def new_session():
        # log.critical('Create new DB session')
        if DatabaseSessionManager.in_flask_request():

            g.session = Database().session_a

            # g.session.refresh()
            # print('new flask session')
        else:
            db.session = Database().session_b

    @staticmethod
    def clear_session():
        if DatabaseSessionManager.in_flask_request():
            # print(f"gsession: {g.session}")
            g.session.remove()
            # g.session = None
        else:
            if db.session:
                db.session.remove()
                db.session = None
            else:
                print('No DB session found to clear!')


class ModelBase:
    """Declarative base that defines default attributes."""
    _hidden_attributes = []

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    # Primary key, internal use only
    id = Column(Integer, primary_key=True)

    @classmethod
    def get(cls, id_=None):

        session = DatabaseSessionManager.get_session()

        result = None

        if id_ is None:
            result = session.query(cls).all()
        else:
            try:
                result = session.query(cls).filter_by(id=id_).one()
            except NoResultFound:
                result = None

        return result

    def save(self) -> None:

        session = DatabaseSessionManager.get_session()

        # new objects do not have an `id`
        if not self.id:
            session.add(self)

        session.commit()

    def delete(self) -> None:

        session = DatabaseSessionManager.get_session()

        session.delete(self)
        session.commit()

    @classmethod
    def help(cls) -> str:
        i = inspect(cls)
        properties = ''.join([f' ->{a.key}\n' for a in i.mapper.column_attrs])
        relations = ''.join([f' ->{a[0]}\n' for a in i.relationships.items()])
        methods = class_inspect.getmembers(cls,
                                           predicate=class_inspect.isroutine)

        methods = ''.join([f' ->{key[0]}\n' for key in methods
                          if not key[0].startswith('_')])

        print(
            f'Table: {cls.__tablename__}\n\n'
            f'Properties: \n{properties}\n'
            f'Relations: \n{relations}\n'
            f'Methods: \n{methods}\n'
        )


Base = declarative_base(cls=ModelBase)



import uuid
import logging

import vantage6.server.model as db

module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


# TODO maybe move this function to a more general location (utils?)
def _is_valid_uuid(uuid_to_test):
    """
    Check if uuid_to_test is a valid UUID.

     Parameters
    ----------
    uuid_to_test : str

     Returns
    -------
    `True` if uuid_to_test is a valid UUID, otherwise `False`.
    """
    try:
        uuid_obj = uuid.UUID(uuid_to_test)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_to_test


def load(fixtures, drop_all=False):
    # TODO we are not sure the DB is connected here....

    # if drop_all:
    #     Database().drop_all()

    log.info("Create Organizations and Users")
    for org in fixtures.get("organizations", {}):
        # print(org)

        # create organization
        organization = db.Organization(**{k: org[k] for k in [
            "name", "domain", "address1", "address2", "zipcode",
            "country", "public_key"
        ]})
        organization.save()
        log.debug(f"processed organization={organization.name}")
        superuserrole = db.Role(name="super", description="Super user",
                                rules=db.Rule.get(), organization=organization)
        # create users
        for usr in org.get("users", {}):
            user = db.User(**usr)
            user.roles = [superuserrole]
            user.organization = organization
            user.save()
            log.debug(f"processed user={user.username}")

    log.info("Create collaborations")
    for col in fixtures.get("collaborations", {}):

        # create collaboration
        collaboration = db.Collaboration(name=col.get("name"), encrypted=col.get("encrypted", True))
        log.debug(f"processed collaboration={collaboration.name}")

        # append organizations to the collaboration

        for participant in col.get("participants", {}):
            if not isinstance(participant, dict) \
                    or not participant.get('name') \
                    or not participant.get('api_key'):
                log.error("Collaboration participants should contain the "
                          "fields 'name' and 'api_key'. This is not the case "
                          f"for participant {participant} in collaboration "
                          f"{collaboration.name}")
                exit(1)

            org_name = participant.get("name")
            node_api_key = participant.get("api_key")

            # check if api key is valid uuid
            if not _is_valid_uuid(node_api_key):
                log.error(f"API key '{node_api_key}' is not a valid UUID!")
                exit(1)

            organization = db.Organization.get_by_name(org_name)
            collaboration.organizations.append(organization)
            log.debug(f"added {org_name} to the collaboration")

            node = db.Node(
                organization=organization,
                collaboration=collaboration,
                name=f"{organization.name} - {collaboration.name} Node",
                api_key=node_api_key
            )
            node.save()
            log.debug(f"added node {node.name} to {collaboration.name}")

        collaboration.save()

        # append dummy tasks to the collaboration
        log.debug("Processing Task Assignments")
        for image in col.get("tasks", {}):
            initiator = collaboration.organizations[0]
            task = db.Task(
                name=f"Example task",
                image=image,
                collaboration=collaboration,
                run_id=db.Task.next_run_id(),
                initiator=initiator
            )

            for organization in collaboration.organizations:
                result = db.Result(
                    task=task,
                    input="something",
                    organization=organization
                )
                result.save()

            task.save()
            log.debug(f"Processed task {task.name}")



# -*- coding: utf-8 -*-
import logging
import json

from flask import g, request, url_for
from http import HTTPStatus
from sqlalchemy import desc

from vantage6.common.globals import STRING_ENCODING
from vantage6.server import db
from vantage6.server.model.base import DatabaseSessionManager
from vantage6.server.permission import (
    Scope as S,
    PermissionManager,
    Operation as P
)
from vantage6.server.resource import only_for, ServicesResources, with_user
from vantage6.server.resource._schema import (
    TaskSchema,
    TaskIncludedSchema,
    TaskResultSchema
)
from vantage6.server.resource.pagination import Pagination


module_name = __name__.split('.')[-1]
log = logging.getLogger(module_name)


def setup(api, api_base, services):
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Tasks,
        path,
        endpoint='task_without_id',
        methods=('GET', 'POST'),
        resource_class_kwargs=services
    )
    api.add_resource(
        Task,
        path + '/<int:id>',
        endpoint='task_with_id',
        methods=('GET', 'DELETE'),
        resource_class_kwargs=services
    )
    api.add_resource(
        TaskResult,
        path + '/<int:id>/result',
        endpoint='task_result',
        methods=('GET',),
        resource_class_kwargs=services
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager):
    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW,
        description="view any task")
    add(scope=S.ORGANIZATION, operation=P.VIEW, assign_to_container=True,
        assign_to_node=True, description="view tasks of your organization")

    add(scope=S.GLOBAL, operation=P.CREATE, description="create a new task")
    add(scope=S.ORGANIZATION, operation=P.CREATE,
        description=(
            "create a new task for collaborations in which your organization "
            "participates with"
        ))

    add(scope=S.GLOBAL, operation=P.DELETE,
        description="delete a task")
    add(scope=S.ORGANIZATION, operation=P.DELETE,
        description=(
            "delete a task from a collaboration in which your organization "
            "participates with"
        ))


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
task_schema = TaskSchema()
task_result_schema = TaskIncludedSchema()
task_result_schema2 = TaskResultSchema()


class TaskBase(ServicesResources):

    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)
        self.r = getattr(self.permissions, module_name)


class Tasks(TaskBase):

    @only_for(['user', 'node', 'container'])
    def get(self):
        """List tasks
        ---
        description: >-
          Returns a list of tasks.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|View|❌|❌|View any task|\n
          |Task|Organization|View|✅|✅|View any task in your organization|
          \n

          Accessible to users.

        parameters:
          - in: query
            name: initiator_id
            schema:
              type: int
            description: The organization id of the origin of the request
          - in: query
            name: collaboration_id
            schema:
              type: int
            description: The collaboration id to which the task belongs
          - in: query
            name: image
            schema:
              type: str
            description: >-
              (Docker) image name which is used in the task. Name to match
              with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: parent_id
            schema:
              type: int
            description: The id of the parent task
          - in: query
            name: run_id
            schema:
              type: int
            description: The run id that belongs to the task
          - in: query
            name: name
            schema:
              type: str
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: description
            schema:
              type: string
            description: >-
              Description to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: database
            schema:
              type: string
            description: >-
              Database description to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: result_id
            schema:
              type: int
            description: A result id that belongs to the task
          - in: query
            name: include
            schema:
              type: array
              items:
                type: string
            description: Include 'results' to get task results. Include
              'metadata' to get pagination metadata. Note that this will
              put the actual data in an envelope.
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized

        security:
            - bearerAuth: []

        tags: ["Task"]
        """
        q = DatabaseSessionManager.get_session().query(db.Task)
        args = request.args

        # obtain organization id
        auth_org_id = self.obtain_organization_id()

        # check permissions and apply filter if neccassary
        if not self.r.v_glo.can():
            if self.r.v_org.can():
                q = q.join(db.Collaboration).join(db.Organization)\
                    .filter(db.Collaboration.organizations.any(id=auth_org_id))
            else:
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        # filter based on arguments
        for param in ['initiator_id', 'collaboration_id', 'parent_id',
                      'run_id']:
            if param in args:
                q = q.filter(getattr(db.Task, param) == args[param])
        for param in ['name', 'image', 'description', 'database']:
            if param in args:
                q = q.filter(getattr(db.Task, param).like(args[param]))
        if 'result_id' in args:
            q = q.join(db.Result).filter(db.Result.id == args['result_id'])

        q = q.order_by(desc(db.Task.id))
        # paginate tasks
        page = Pagination.from_query(q, request)

        # serialization schema
        schema = task_result_schema if self.is_included('result') else\
            task_schema

        return self.response(page, schema)

    @only_for(["user", "container"])
    def post(self):
        """Adds new computation task
        ---
        description: >-
          Creates a new task within a collaboration. If no `organization_ids`
          are given the task is send to all organizations within the
          collaboration. The endpoint can be accessed by both a `User` and
          `Container`.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|Create|❌|❌|Create a new task|\n
          |Task|Organization|Create|❌|✅|Create a new task for a specific
          collaboration in which your organization participates|\n

          ## Accessed as `User`\n
          This endpoint is accessible to users. A new `run_id` is
          created when a user creates a task. The user needs to be within an
          organization that is part of the collaboration to which the task is
          posted.\n

          ## Accessed as `Container`\n
          When this endpoint is accessed by an algorithm container, it is
          considered to be a child-task of the container, and will get the
          `run_id` from the initial task. Containers have limited permissions
          to create tasks: they are only allowed to create tasks in the same
          collaboration using the same image.\n

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Task'

        responses:
          200:
            description: Ok
          400:
            description: Supplied organizations are not in the supplied
              collaboration, or not all required nodes are registered, or you
              are not in the collaboration yourself
          401:
            description: Unauthorized
          404:
            description: Collaboration with `collaboration_id` not found

        security:
          - bearerAuth: []

        tags: ["Task"]
        """
        data = request.get_json()
        collaboration_id = data.get('collaboration_id')
        collaboration = db.Collaboration.get(collaboration_id)

        if not collaboration:
            return {"msg": f"Collaboration id={collaboration_id} not found!"},\
                   HTTPStatus.NOT_FOUND

        organizations_json_list = data.get('organizations')
        org_ids = [org.get("id") for org in organizations_json_list]
        db_ids = collaboration.get_organization_ids()

        # Check that all organization ids are within the collaboration, this
        # also ensures us that the organizations exist
        if not set(org_ids).issubset(db_ids):
            return {"msg": (
                "At least one of the supplied organizations in not within "
                "the collaboration."
            )}, HTTPStatus.BAD_REQUEST

        # check if all the organizations have a registered node
        nodes = DatabaseSessionManager.get_session().query(db.Node)\
            .filter(db.Node.organization_id.in_(org_ids))\
            .filter(db.Node.collaboration_id == collaboration_id)\
            .all()
        if len(nodes) < len(org_ids):
            present_nodes = [node.organization_id for node in nodes]
            missing = [str(id) for id in org_ids if id not in present_nodes]
            return {"msg": (
                "Cannot create this task because there are no nodes registered"
                f" for the following organization(s): {', '.join(missing)}."
            )}, HTTPStatus.BAD_REQUEST

        # figure out the initiator organization of the task
        if g.user:
            initiator = g.user.organization
        else:  # g.container:
            initiator = db.Node.get(g.container["node_id"]).organization

        # check if the initiator is part of the collaboration
        if initiator not in collaboration.organizations:
            return {
                "msg": "You can only create tasks for collaborations "
                       "you are participating in!"
            }, HTTPStatus.UNAUTHORIZED

        # Create the new task in the database
        image = data.get('image', '')

        # verify permissions
        if g.user:
            if not self.r.c_glo.can():
                c_orgs = collaboration.organizations
                if not (self.r.c_org.can() and g.user.organization in c_orgs):
                    return {'msg': 'You lack the permission to do that!'}, \
                        HTTPStatus.UNAUTHORIZED

        elif g.container:
            # verify that the container has permissions to create the task
            if not self.__verify_container_permissions(g.container, image,
                                                       collaboration_id):
                return {"msg": "Container-token is not valid"}, \
                    HTTPStatus.UNAUTHORIZED

        # permissions ok, create record
        task = db.Task(collaboration=collaboration, name=data.get('name', ''),
                       description=data.get('description', ''), image=image,
                       database=data.get('database', ''), initiator=initiator)

        # create run_id. Users can only create top-level -tasks (they will not
        # have sub-tasks). Therefore, always create a new run_id. Tasks created
        # by containers are always sub-tasks
        if g.user:
            task.run_id = task.next_run_id()
            log.debug(f"New run_id {task.run_id}")
        elif g.container:
            task.parent_id = g.container["task_id"]
            task.run_id = db.Task.get(g.container["task_id"]).run_id
            log.debug(f"Sub task from parent_id={task.parent_id}")

        # ok commit session...
        task.save()

        # if the 'master'-flag is set to true the (master) task is executed on
        # a node in the collaboration from the organization to which the user
        # belongs. If also organization_ids are supplied, then these are
        # ignored.
        # TODO in this case the user *must* have a node attached to this
        # collaboration
        # TODO this does not make a lot of sense as the `organizations` input
        # should only contain the organization where the master container
        # shoudl run
        assign_orgs = []
        if data.get("master", False) and g.user:
            for org in organizations_json_list:
                if org['id'] == g.user.organization_id:
                    assign_orgs = [org]
                    break
            if not assign_orgs:
                return {'msg': 'You\'re trying to create a master task. '
                        'However you do not have a node yourself in this '
                        'collaboration!'}, HTTPStatus.BAD_REQUEST
        else:
            assign_orgs = organizations_json_list

        # now we need to create results for the nodes to fill. Each node
        # receives their instructions from a result, not from the task itself
        log.debug(f"Assigning task to {len(assign_orgs)} nodes.")
        for org in assign_orgs:
            organization = db.Organization.get(org['id'])
            log.debug(f"Assigning task to '{organization.name}'.")
            input_ = org.get('input')
            # FIXME: legacy input from the client, could be removed at some
            # point
            if isinstance(input_, dict):
                input_ = json.dumps(input_).encode(STRING_ENCODING)
            # Create result
            result = db.Result(
                task=task,
                organization=organization,
                input=input_,
            )
            result.save()

        # notify nodes a new task available (only to online nodes), nodes that
        # are offline will receive this task on sign in.
        self.socketio.emit('new_task', task.id, namespace='/tasks',
                           room=f'collaboration_{task.collaboration_id}')

        # add some logging
        log.info(f"New task for collaboration '{task.collaboration.name}'")
        if g.user:
            log.debug(f" created by: '{g.user.username}'")
        else:
            log.debug(f" created by container on node_id="
                      f"{g.container['node_id']}"
                      f" for (master) task_id={g.container['task_id']}")

        log.debug(f" url: '{url_for('task_with_id', id=task.id)}'")
        log.debug(f" name: '{task.name}'")
        log.debug(f" image: '{task.image}'")

        return task_schema.dump(task, many=False).data, HTTPStatus.CREATED

    @staticmethod
    def __verify_container_permissions(container, image, collaboration_id):
        """Validates that the container is allowed to create the task."""

        # check that the image is allowed
        # if container["image"] != task.image:
        # FIXME why?
        if not image.endswith(container["image"]):
            log.warning((f"Container from node={container['node_id']} "
                        f"attempts to post a task using illegal image!?"))
            log.warning(f"  task image: {image}")
            log.warning(f"  container image: {container['image']}")
            return False

        # check master task is not completed yet
        if db.Task.get(container["task_id"]).complete:
            log.warning(
                f"Container from node={container['node_id']} "
                f"attempts to start sub-task for a completed "
                f"task={container['task_id']}"
            )
            return False

        # check that node id is indeed part of the collaboration
        if not container["collaboration_id"] == collaboration_id:
            log.warning(
                f"Container attempts to create a task outside "
                f"collaboration_id={container['collaboration_id']} in "
                f"collaboration_id={collaboration_id}!"
            )
            return False

        return True


class Task(TaskBase):
    """Resource for /api/task"""

    @only_for(["user", "node", "container"])
    def get(self, id):
        """Get task
        ---
        description: >-
          Returns the task specified by the id.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|View|❌|❌|View any task|\n
          |Task|Organization|View|✅|✅|View any task in your organization|

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Task id
            required: true
          - in: query
            name: include
            schema:
              type: string
            description: Include 'results' to include the task's results.

        responses:
          200:
            description: Ok
          404:
            description: Task not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Task"]
        """
        task = db.Task.get(id)
        if not task:
            return {"msg": f"task id={id} is not found"}, HTTPStatus.NOT_FOUND

        # determine the organization to which the auth belongs
        auth_org = self.obtain_auth_organization()

        # obtain schema
        schema = task_result_schema if request.args.get('include') == \
            'results' else task_schema

        # check permissions
        if not self.r.v_glo.can():
            org_ids = [org.id for org in task.collaboration.organizations]
            if not (self.r.v_org.can() and auth_org.id in org_ids):
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        return schema.dump(task, many=False).data, HTTPStatus.OK

    @with_user
    def delete(self, id):
        """Remove task
        ---
        description: >-
          Remove tasks and their results.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|Delete|❌|❌|Delete a task|\n
          |Task|Organization|Delete|❌|❌|Delete a task from a collaboration
          in which your organization participates|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Task id
            required: true

        responses:
          200:
            description: Ok
          404:
            description: Task not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Task"]
        """

        task = db.Task.get(id)
        if not task:
            return {"msg": f"task id={id} not found"}, HTTPStatus.NOT_FOUND

        # validate permissions
        if not self.r.d_glo.can():
            orgs = task.collaboration.organizations
            if not (self.r.d_org.can() and g.user.organization in orgs):
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        # retrieve results that belong to this task
        log.info(f'Removing task id={task.id}')
        for result in task.results:
            log.info(f" Removing result id={result.id}")
            result.delete()

        # permissions ok, delete...
        task.delete()

        return {"msg": f"task id={id} and its result successfully deleted"}, \
            HTTPStatus.OK


class TaskResult(ServicesResources):
    """Resource for /api/task/<int:id>/result"""

    def __init__(self, socketio, mail, api, permissions, config):
        super().__init__(socketio, mail, api, permissions, config)
        self.r = getattr(self.permissions, "result")

    @only_for(['user', 'container'])
    def get(self, id):
        """Return the results for a specific task
        ---
        description: >-
          Returns the task's results specified by the task id.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Result|Global|View|❌|❌|View any result|\n
          |Result|Organization|View|✅|✅|View results for the
          collaborations in which your organization participates with|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Task id
            required: true
          - in: query
            name: include
            schema:
              type: string
            description: Include 'metadata' to get pagination metadata. Note
              that this will put the actual data in an envelope.
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page

        responses:
          200:
            description: Ok
          404:
            description: Task not found
          401:
            description: Unauthorized

        security:
            - bearerAuth: []

        tags: ["Task"]
        """
        task = db.Task.get(id)
        if not task:
            return {"msg": f"task id={id} not found"}, \
                HTTPStatus.NOT_FOUND

        # obtain organization model
        org = self.obtain_auth_organization()

        if not self.r.v_glo.can():
            c_orgs = task.collaboration.organizations
            if not (self.r.v_org.can() and org in c_orgs):
                return {'msg': 'You lack the permission to do that!'}, \
                    HTTPStatus.UNAUTHORIZED

        # pagination
        page = Pagination.from_list(task.results, request)

        # model serialization
        return self.response(page, task_result_schema2)



# -*- coding: utf-8 -*-
import datetime
from uuid import uuid1
import yaml
import unittest
import logging
import json
import uuid

from http import HTTPStatus
from unittest.mock import patch
from flask import Response as BaseResponse
from flask.testing import FlaskClient
from werkzeug.utils import cached_property

from vantage6.common import logger_name
from vantage6.common.globals import APPNAME
from vantage6.server.globals import PACAKAGE_FOLDER
from vantage6.server import ServerApp, db
from vantage6.server.model import (Rule, Role, Organization, User, Node,
                                   Collaboration, Task, Result)
from vantage6.server.model.rule import Scope, Operation
from vantage6.server import context
from vantage6.server._version import __version__
from vantage6.server.model.base import Database, DatabaseSessionManager
from vantage6.server.controller.fixture import load


logger = logger_name(__name__)
log = logging.getLogger(logger)


class Response(BaseResponse):
    @cached_property
    def json(self):
        return json.loads(self.data)


class TestNode(FlaskClient):
    def open(self, *args, **kwargs):
        if 'json' in kwargs:
            kwargs['data'] = json.dumps(kwargs.pop('json'))
            kwargs['content_type'] = 'application/json'
        return super().open(*args, **kwargs)


class TestResources(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Called immediately before running a test method."""
        Database().connect("sqlite://", allow_drop_all=True)

        ctx = context.TestContext.from_external_config_file(
            "unittest_config.yaml")

        server = ServerApp(ctx)
        cls.server = server

        file_ = str(PACAKAGE_FOLDER / APPNAME / "server" / "_data" /
                    "unittest_fixtures.yaml")
        with open(file_) as f:
            cls.entities = yaml.safe_load(f.read())
        load(cls.entities)

        server.app.testing = True
        cls.app = server.app.test_client()

        cls.credentials = {
            'root': {
                'username': 'root',
                'password': 'root'
            },
            'admin': {
                'username': 'frank@iknl.nl',
                'password': 'password'
            },
            'user': {
                'username': 'melle@iknl.nl',
                'password': 'password'
            },
            'user-to-delete': {
                'username': 'dont-use-me',
                'password': 'password'
            }
        }

    @classmethod
    def tearDownClass(cls):
        Database().clear_data()

    @classmethod
    def setUp(cls):
        # set db.session
        DatabaseSessionManager.get_session()

    @classmethod
    def tearDown(cls):
        # unset db.session
        DatabaseSessionManager.clear_session()

    def login(self, type_='root'):
        with self.server.app.test_client() as client:
            tokens = client.post(
                '/api/token/user',
                json=self.credentials[type_]
            ).json
        if 'access_token' in tokens:
            headers = {
                'Authorization': 'Bearer {}'.format(tokens['access_token'])
            }
            return headers
        else:
            print('something wrong, during login:')
            print(tokens)
            return None

    def create_user(self, organization=None, rules=[], password="password"):

        if not organization:
            organization = Organization(name="some-organization")
            organization.save()

        # user details
        username = str(uuid.uuid1())

        # create a temporary organization
        user = User(username=username, password=password,
                    organization=organization, email=f"{username}@test.org",
                    rules=rules)
        user.save()

        self.credentials[username] = {
            "username": username,
            "password": password
        }

        return user

    def create_node(self, organization=None, collaboration=None):
        if not organization:
            organization = Organization()

        if not collaboration:
            collaboration = Collaboration()

        api_key = str(uuid1())
        node = Node(
            name=str(uuid1()),
            api_key=api_key,
            organization=organization,
            collaboration=collaboration
        )
        node.save()

        return node, api_key

    def login_node(self, api_key):
        tokens = self.app.post(
            '/api/token/node',
            json={"api_key": api_key}
        ).json
        if 'access_token' in tokens:
            headers = {
                'Authorization': 'Bearer {}'.format(tokens['access_token'])
            }
        else:
            print(tokens)

        return headers

    def login_container(self, collaboration=None, organization=None,
                        node=None, task=None, api_key=None):
        if not node:
            if not collaboration:
                collaboration = Collaboration()
            if not organization:
                organization = Organization()
            api_key = str(uuid1())
            node = Node(organization=organization, collaboration=collaboration,
                        api_key=api_key)
            node.save()
        else:
            collaboration = node.collaboration
            organization = node.organization

        if not task:
            task = Task(image="some-image", collaboration=collaboration,
                        results=[Result()])
            task.save()

        headers = self.login_node(api_key)
        tokens = self.app.post('/api/token/container', headers=headers, json={
            "image": "some-image",
            "task_id": task.id
        }
        ).json

        if 'msg' in tokens:
            print(tokens['msg'])

        headers = {
            'Authorization': 'Bearer {}'.format(tokens['container_token'])
        }
        return headers

    def create_node_and_login(self, *args, **kwargs):
        node, api_key = self.create_node(*args, **kwargs)
        return self.login_node(api_key)

    def create_user_and_login(self, organization=None, rules=[]):
        user = self.create_user(organization, rules)
        return self.login(user.username)

    def test_version(self):
        rv = self.app.get('/api/version')
        r = json.loads(rv.data)
        self.assertIn('version', r)
        self.assertEqual(r['version'], __version__)

    def test_token_different_users(self):
        for type_ in ["root", "admin", "user"]:
            tokens = self.app.post(
                '/api/token/user',
                json=self.credentials[type_]
            ).json
            self.assertIn('access_token', tokens)
            self.assertIn('refresh_token', tokens)

    def test_organization(self):

        rule = Rule.get_by_("organization", Scope.GLOBAL,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])

        # First retrieve a list of all organizations
        orgs = self.app.get('/api/organization', headers=headers).json
        self.assertEqual(len(orgs), len(Organization.get()))

        attrs = [
            'id',
            'name',
            'domain',
            'name',
            'address1',
            'address2',
            'zipcode',
            'country',
        ]

        org = orgs[0]
        for attr in attrs:
            self.assertIn(attr, org)

        # Retrieve a single organization
        url = f'/api/organization/{org["id"]}'
        org = self.app.get(url, headers=headers).json
        self.assertEqual(org['id'], orgs[0]['id'])
        self.assertEqual(org['name'], orgs[0]['name'])

        # Create a new organization
        org_details = {
            'name': 'Umbrella Corporation',
            'address1': 'Resident Evil Pike'
        }

        org = self.app.post(
            '/api/organization',
            json=org_details,
            headers=headers
        ).json

        # for attr in attrs:
        #     self.assertIn(attr, org)

        # self.assertGreater(org['id'], 0)

        orgs = self.app.get('/api/organization', headers=headers).json
        # self.assertEqual(len(orgs), 4)

    def test_collaboration(self):
        org = Organization()
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(organization=org, rules=[rule])

        collaborations = self.app.get(
            '/api/collaboration', headers=headers
        )
        self.assertEqual(collaborations.status_code, HTTPStatus.OK)
        db_cols = Collaboration.get()
        self.assertEqual(len(collaborations.json), len(db_cols))

    def test_node_without_id(self):

        # GET
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        nodes = self.app.get("/api/node", headers=headers).json
        expected_fields = [
            'name',
            'collaboration',
            'organization',
            'status',
            'id',
            'type',
            'last_seen',
            'ip'
        ]
        for node in nodes:
            for key in expected_fields:
                self.assertIn(key, node)

        nodes = self.app.get("/api/node", headers=headers).json
        self.assertIsNotNone(nodes)

        # POST
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.CREATE)
        headers = self.create_user_and_login(rules=[rule])
        # unknown collaboration id should fail
        response = self.app.post("/api/node", headers=headers, json={
            "collaboration_id": 99999
        })
        response_json = response.json
        self.assertIn("msg", response_json)
        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)

        # succesfully create a node
        org = Organization()
        col = Collaboration(organizations=[org])
        col.save()

        headers = self.create_user_and_login(org, rules=[rule])
        response = self.app.post("/api/node", headers=headers, json={
            "collaboration_id": col.id
        })
        self.assertEqual(response.status_code, HTTPStatus.CREATED)

    def test_node_with_id(self):

        # root user can access all nodes
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        node = self.app.get("/api/node/8", headers=headers).json
        expected_fields = [
            'name',
            'collaboration',
            'organization',
            'status',
            'id',
            'type',
            'last_seen',
            'ip'
        ]
        for key in expected_fields:
            self.assertIn(key, node)

        # user cannot access all
        headers = self.create_user_and_login()
        node = self.app.get("/api/node/8", headers=headers)
        self.assertEqual(node.status_code, HTTPStatus.UNAUTHORIZED)

        # some nodes just don't exist
        node = self.app.get("/api/node/9999", headers=headers)
        self.assertEqual(node.status_code, 404)

    def test_result_with_id(self):
        headers = self.login("root")
        result = self.app.get("/api/result/1", headers=headers)
        self.assertEqual(result.status_code, 200)

        result = self.app.get("/api/result/1?include=task", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_result_without_id(self):
        headers = self.login("root")
        result1 = self.app.get("/api/result", headers=headers)
        self.assertEqual(result1.status_code, 200)

        result2 = self.app.get("/api/result?state=open&&node_id=1",
                               headers=headers)
        self.assertEqual(result2.status_code, 200)

        result3 = self.app.get("/api/result?task_id=1", headers=headers)
        self.assertEqual(result3.status_code, 200)

        result4 = self.app.get("/api/result?task_id=1&&node_id=1",
                               headers=headers)
        self.assertEqual(result4.status_code, 200)

    def test_stats(self):
        headers = self.login("root")
        result = self.app.get("/api/result", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_with_id(self):
        headers = self.login("root")
        result = self.app.get("/api/task/1", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_witout_id(self):
        headers = self.login("root")
        result = self.app.get("/api/task", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_including_results(self):
        headers = self.login("root")
        result = self.app.get("/api/task?include=results", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_task_unknown(self):
        headers = self.login("root")
        result = self.app.get("/api/task/9999", headers=headers)
        self.assertEqual(result.status_code, 404)

    def test_user_with_id(self):
        headers = self.login("admin")
        result = self.app.get("/api/user/1", headers=headers)
        self.assertEqual(result.status_code, 200)
        user = result.json

        expected_fields = [
            "username",
            "firstname",
            "lastname",
            "roles"
        ]
        for field in expected_fields:
            self.assertIn(field, user)

    def test_user_unknown(self):
        headers = self.login("admin")
        result = self.app.get("/api/user/9999", headers=headers)
        self.assertEqual(result.status_code, 404)

    def test_user_without_id(self):
        for role in ["user", "admin", "root"]:
            headers = self.login(role)
            result = self.app.get("/api/user", headers=headers)
            self.assertEqual(result.status_code, 200)

    def test_user_post(self):
        headers = self.login("root")
        new_user = {
            "username": "unittest",
            "firstname": "unit",
            "lastname": "test",
            "email": "unit@test.org",
        }
        # with a bad password, user should not be created
        new_user['password'] = "1234"
        result = self.app.post('/api/user', headers=headers, json=new_user)
        self.assertEqual(result.status_code, HTTPStatus.BAD_REQUEST)

        new_user['password'] = "Welkom01!"
        result = self.app.post("/api/user", headers=headers,
                               json=new_user)
        self.assertEqual(result.status_code, 201)

        result = self.app.post("/api/user", headers=headers,
                               json=new_user)
        self.assertEqual(result.status_code, 400)

    def test_user_delete(self):
        headers = self.login("root")
        result = self.app.delete("/api/user/5", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_user_delete_unknown(self):
        headers = self.login("root")
        result = self.app.delete("/api/user/99999", headers=headers)
        self.assertEqual(result.status_code, 404)

    def test_user_patch(self):
        headers = self.login("root")
        result = self.app.patch("/api/user/2", headers=headers, json={
            "firstname": "Henk",
            "lastname": "Martin"
        })
        self.assertEqual(result.status_code, 200)

    def test_user_patch_unknown(self):
        headers = self.login("root")
        result = self.app.patch("/api/user/9999", headers=headers, json={
            "username": "root2"
        })
        self.assertEqual(result.status_code, 404)

    def test_root_role_forbidden(self):
        headers = self.login("root")
        new_user = {
            "username": "some",
            "firstname": "guy",
            "lastname": "there",
            "roles":  "root",
            "password": "super-secret"
        }
        result = self.app.post("/api/user", headers=headers,
                               json=new_user)
        self.assertEqual(result.status_code, 400)

    @patch("vantage6.server.mail_service.MailService.send_email")
    def test_reset_password(self, send_email):
        user_ = {
            "username": "root"
        }
        result = self.app.post("/api/recover/lost", json=user_)
        self.assertEqual(result.status_code, 200)

    @patch("vantage6.server.mail_service.MailService.send_email")
    def test_reset_password_missing_error(self, send_email):
        result = self.app.post("/api/recover/lost", json={})
        self.assertEqual(result.status_code, 400)

    @patch("vantage6.server.resource.recover.decode_token")
    def test_recover_password(self, decode_token):
        decode_token.return_value = {'sub': {'id': 1}}
        new_password = {
            "password": "$Ecret88!",
            "reset_token": "token"
        }
        result = self.app.post("/api/recover/reset", json=new_password)
        self.assertEqual(result.status_code, 200)

        # verify that the new password works
        result = self.app.post("/api/token/user", json={
            "username": "root",
            "password": "$Ecret88!"
        })
        self.assertIn("access_token", result.json)
        self.credentials["root"]["password"] = "$Ecret88!"

    def test_fail_recover_password(self):
        result = self.app.post("/api/recover/reset", json={})
        self.assertEqual(result.status_code, 400)

    def test_change_password(self):
        user = self.create_user(password="Password1!")
        headers = self.login(user.username)

        # test if fails when not providing correct data
        result = self.app.patch("/api/password/change", headers=headers, json={
            "current_password": "Password1!"
        })
        self.assertEqual(result.status_code, 400)
        result = self.app.patch("/api/password/change", headers=headers, json={
            "new_password": "a_new_password"
        })
        self.assertEqual(result.status_code, 400)

        # test if fails when wrong password is provided
        result = self.app.patch("/api/password/change", headers=headers, json={
            "current_password": "wrong_password1!",
            "new_password": "a_new_password"
        })
        self.assertEqual(result.status_code, 401)

        # test if fails when new password is the same
        result = self.app.patch("/api/password/change", headers=headers, json={
            "current_password": "Password1!",
            "new_password": "Password1!"
        })
        self.assertEqual(result.status_code, 400)

        # test if it works when used as intended
        result = self.app.patch("/api/password/change", headers=headers, json={
            "current_password": "Password1!",
            "new_password": "A_new_password1"
        })
        self.assertEqual(result.status_code, 200)
        db.session.refresh(user)
        self.assertTrue(user.check_password("A_new_password1"))

    def test_view_rules(self):
        headers = self.login("root")
        result = self.app.get("/api/rule", headers=headers)
        self.assertEqual(result.status_code, 200)

    def test_view_roles(self):
        headers = self.login("root")
        result = self.app.get("/api/role", headers=headers)
        self.assertEqual(result.status_code, 200)

        body = result.json
        expected_fields = ['organization', 'name', 'description', 'users']
        for field in expected_fields:
            self.assertIn(field, body[0])

    def test_create_role_as_root(self):
        headers = self.login("root")

        # obtain available rules
        rules = self.app.get("/api/rule", headers=headers).json
        rule_ids = [rule.get("id") for rule in rules]

        # assign first two rules to role
        body = {
            "name": "some-role-name",
            "description": "Testing if we can create a role",
            "rules": rule_ids[:2]
        }

        # create role
        result = self.app.post("/api/role", headers=headers, json=body)

        # check that server responded ok
        self.assertEqual(result.status_code, HTTPStatus.CREATED)

        # verify the values
        self.assertEqual(result.json.get("name"), body["name"])
        self.assertEqual(result.json.get("description"), body["description"])
        self.assertEqual(len(result.json.get("rules")), 2)

    def test_create_role_as_root_for_different_organization(self):
        headers = self.login("root")

        # obtain available rules
        rules = self.app.get("/api/rule", headers=headers).json

        # create new organization, so we're sure that the current user
        # is not assigned to the same organization
        org = Organization(name="Some-random-organization")
        org.save()

        body = {
            "name": "some-role-name",
            "description": "Testing if we can create a rol for another org",
            "rules": [rule.get("id") for rule in rules],
            "organization_id": org.id
        }

        # create role
        result = self.app.post("/api/role", headers=headers, json=body)

        # check that server responded ok
        self.assertEqual(result.status_code, HTTPStatus.CREATED)

        # verify the organization
        self.assertEqual(org.id, result.json["organization"]["id"])

    def test_create_role_permissions(self):
        all_rules = Rule.get()

        # check user without any permissions
        headers = self.create_user_and_login()

        body = {
            "name": "some-role-name",
            "description": "Testing if we can create a rol for another org",
            "rules": [rule.id for rule in all_rules],
        }
        result = self.app.post("/api/role", headers=headers, json=body)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # check that user with a missing rule cannot create a role with that
        # missing rule
        headers = self.create_user_and_login(rules=(all_rules[:-2]))
        result = self.app.post("/api/role", headers=headers, json=body)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # check that user can create role within his organization
        rule = Rule.get_by_("role", scope=Scope.ORGANIZATION,
                            operation=Operation.CREATE)

        headers = self.create_user_and_login(rules=[rule])
        body["rules"] = [rule.id]
        result = self.app.post("/api/role", headers=headers, json=body)

        self.assertEqual(result.status_code, HTTPStatus.CREATED)

        # check a non-existing organization
        headers = self.login("root")
        body["organization_id"] = -1
        result = self.app.post('/api/role', headers=headers, json=body)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # check that assigning an unexisting rule is not possible
        headers = self.create_user_and_login()
        body["rules"] = [-1]
        result = self.app.post("/api/role", headers=headers, json=body)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

    def test_edit_role(self):
        headers = self.login('root')

        # create testing entities
        org = Organization(name="some-organization-name")
        org.save()
        role = Role(name="some-role-name", organization=org)
        role.save()

        # test name, description
        result = self.app.patch(f'/api/role/{role.id}', headers=headers, json={
            "name": "a-different-role-name",
            "description": "some description of this role..."
        })

        db.session.refresh(role)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(role.name, "a-different-role-name")
        self.assertEqual(role.description, "some description of this role...")

        # test modifying rules
        all_rule_ids = [rule.id for rule in Rule.get()]
        result = self.app.patch(f'/api/role/{role.id}', headers=headers, json={
            "rules": all_rule_ids
        })
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertListEqual(all_rule_ids, [rule.id for rule in role.rules])

        # test non owning rules
        rule = Rule.get_by_("role", Scope.ORGANIZATION,
                            Operation.EDIT)
        headers = self.create_user_and_login(org, [rule])
        result = self.app.patch(f"/api/role/{role.id}", headers=headers, json={
            "rules": all_rule_ids
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test modifying role of another organization, without global
        # permission
        org2 = Organization(name="another-organization")
        headers = self.create_user_and_login(org2, [rule])
        result = self.app.patch(f'/api/role/{role.id}', headers=headers, json={
            "name": "this-will-not-be-updated"
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test modifying role with global permissions
        rule = Rule.get_by_("role", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(org2, [rule])
        result = self.app.patch(f'/api/role/{role.id}', headers=headers, json={
            "name": "this-will-not-be-updated"
        })
        self.assertEqual(result.status_code, HTTPStatus.OK)

    def test_remove_role(self):

        org = Organization()
        org.save()
        role = Role(organization=org)
        role.save()

        # test removal without permissions
        headers = self.create_user_and_login()
        result = self.app.delete(f'/api/role/{role.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test removal with organization permissions
        rule = Rule.get_by_("role", Scope.ORGANIZATION,
                            Operation.DELETE)
        headers = self.create_user_and_login(org, [rule])
        result = self.app.delete(f'/api/role/{role.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        # test failed removal with organization permissions
        role = Role(organization=org)  # because we removed it...
        role.save()
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.delete(f'/api/role/{role.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test removal with global permissions
        rule = Rule.get_by_("role", Scope.GLOBAL, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.delete(f'/api/role/{role.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

    def test_rules_from_role(self):
        headers = self.login('root')
        role = Role.get()[0]

        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(role.rules), len(result.json))

        result = self.app.get('/api/role/-1/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

    def test_add_single_rule_to_role(self):
        headers = self.login('root')

        role = Role(name="empty", organization=Organization())
        role.save()

        # role without rules
        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)

        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(result.json), 0)

        rule = Rule.get()[0]

        # try to add rule to non existing role
        result = self.app.post(f'/api/role/-1/rule/{rule.id}',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # try to add non existant rule
        result = self.app.post(f'/api/role/{role.id}/rule/-1',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # add a rule to a role
        result = self.app.post(f'/api/role/{role.id}/rule/{rule.id}',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.CREATED)

        # check that the role now has one rule
        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(result.json), 1)

    def test_remove_single_rule_from_role(self):
        headers = self.login('root')

        rule = Rule.get()[0]
        role = Role(name="unit", organization=Organization(), rules=[rule])
        role.save()

        # try to add rule to non existing role
        result = self.app.delete(f'/api/role/-1/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # try to add non existant rule
        result = self.app.delete(f'/api/role/{role.id}/rule/-1',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(result.json), 1)

        result = self.app.delete(f'/api/role/{role.id}/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(result.json), 0)

    def test_view_permission_rules(self):
        rule = Rule.get_by_("role", Scope.ORGANIZATION, Operation.VIEW)

        role = Role(name="some-role", organization=Organization())
        role.save()

        # user does not belong to organization
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # user does belong to the organization
        headers = self.create_user_and_login(organization=role.organization,
                                             rules=[rule])
        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        # user has global permissions
        rule = Rule.get_by_("role", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.get(f'/api/role/{role.id}/rule', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        role.delete()

    def test_add_rule_to_role_permission(self):

        role = Role(name="new-role", organization=Organization())
        role.save()

        rule = Rule.get_by_("role", Scope.ORGANIZATION, Operation.EDIT)

        # try adding a rule without any permission
        headers = self.create_user_and_login()
        result = self.app.post(f'/api/role/{role.id}/rule/{rule.id}',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # you cant edit other organizations roles
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.post(f'/api/role/{role.id}/rule/{rule.id}',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # you can edit other organizations with the global permission
        rule = Rule.get_by_("role", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.post(f'/api/role/{role.id}/rule/{rule.id}',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.CREATED)

        # however you can only assign rules that you own
        rule = Rule.get_by_("role", Scope.ORGANIZATION, Operation.EDIT)
        result = self.app.post(f'/api/role/{role.id}/rule/{rule.id}',
                               headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        role.delete()

    def test_remove_rule_from_role_permissions(self):

        role = Role(name="new-role", organization=Organization())
        role.save()
        rule = Rule.get_by_("role", Scope.ORGANIZATION,
                            Operation.DELETE)

        # try removing without any permissions
        headers = self.create_user_and_login()
        result = self.app.delete(f'/api/role/{role.id}/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # try removing rule from other organization
        headers = self.create_user_and_login(organization=Organization(),
                                             rules=[rule])
        result = self.app.delete(f'/api/role/{role.id}/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # try removing rule which is not in the role
        headers = self.create_user_and_login(organization=role.organization,
                                             rules=[rule])
        result = self.app.delete(f'/api/role/{role.id}/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        role.rules.append(rule)
        role.save()

        # lets try that again
        headers = self.create_user_and_login(organization=role.organization,
                                             rules=[rule])
        result = self.app.delete(f'/api/role/{role.id}/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        role.rules.append(rule)
        role.save()

        # power users can edit other organization rules
        power_rule = Rule.get_by_("role", Scope.GLOBAL,
                                  Operation.DELETE)
        headers = self.create_user_and_login(rules=[power_rule, rule])
        result = self.app.delete(f'/api/role/{role.id}/rule/{rule.id}',
                                 headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        role.delete()

    def test_view_permission_user(self):

        # user not found
        headers = self.create_user_and_login()
        result = self.app.get('/api/user/-1', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # try to view users without any permissions
        headers = self.create_user_and_login()
        result = self.app.get('/api/user', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # root user can view all users
        headers = self.login('root')
        result = self.app.get('/api/user', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(result.json), len(User.get()))

        # view users of your organization
        rule = Rule.get_by_("user", Scope.ORGANIZATION, Operation.VIEW)
        org = Organization.get(1)
        headers = self.create_user_and_login(org, rules=[rule])
        result = self.app.get('/api/user', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(len(result.json), len(org.users))

        # view a single user of your organization
        user_id = org.users[0].id
        result = self.app.get(f'/api/user/{user_id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        # user can view their own data. This should always be possible
        user = self.create_user(rules=[])
        headers = self.login(user.username)
        result = self.app.get(f'/api/user/{user.id}', headers=headers)

    def test_bounce_existing_username_and_email(self):
        headers = self.create_user_and_login()
        User(username="something", email="mail@me.org").save()
        userdata = {
            "username": "not-important",
            "firstname": "name",
            "lastname": "lastname",
            "password": "welkom01",
            "email": "mail@me.org"
        }
        result = self.app.post('/api/user', headers=headers, json=userdata)
        self.assertEqual(result.status_code, HTTPStatus.BAD_REQUEST)

        userdata['username'] = 'not-important'
        result = self.app.post('/api/user', headers=headers, json=userdata)
        self.assertEqual(result.status_code, HTTPStatus.BAD_REQUEST)

    def test_new_permission_user(self):
        userdata = {
            "username": "smarty",
            "firstname": "Smart",
            "lastname": "Pants",
            "password": "Welkom01!",
            "email": "mail-us@me.org"
        }

        # Creating users for other organizations can only be by global scope
        org = Organization()
        rule = Rule.get_by_("user", Scope.ORGANIZATION,
                            Operation.CREATE)
        userdata['organization_id'] = 1
        headers = self.create_user_and_login(org, rules=[rule])
        result = self.app.post('/api/user', headers=headers, json=userdata)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # you can do that when you have the global scope
        gl_rule = Rule.get_by_("user", Scope.GLOBAL, Operation.CREATE)
        userdata['rules'] = [gl_rule.id]
        headers = self.create_user_and_login(org, rules=[gl_rule])
        result = self.app.post('/api/user', headers=headers, json=userdata)
        self.assertEqual(result.status_code, HTTPStatus.CREATED)

        # you need to own all rules in order to assign them
        headers = self.create_user_and_login(org, rules=[rule])
        userdata['username'] = 'smarty2'
        userdata['email'] = 'mail2@me.org'
        result = self.app.post('/api/user', headers=headers, json=userdata)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # you can only assign roles in which you have all rules
        headers = self.create_user_and_login(org, rules=[rule])
        role = Role(rules=[rule], organization=org)
        role.save()
        userdata['username'] = 'smarty3'
        userdata['email'] = 'mail3@me.org'
        userdata['roles'] = [role.id]
        del userdata['organization_id']
        del userdata['rules']
        result = self.app.post('/api/user', headers=headers, json=userdata)
        self.assertEqual(result.status_code, HTTPStatus.CREATED)
        self.assertEqual(len(result.json['roles']), 1)

    def test_patch_user_permissions(self):

        org = Organization()
        user = User(firstname="Firstname", lastname="Lastname",
                    username="Username", password="Password", email="a@b.c",
                    organization=org)
        user.save()
        self.credentials[user.username] = {'username': user.username,
                                           'password': "Password"}

        # check non-existing user
        headers = self.create_user_and_login()
        result = self.app.patch('/api/user/-1', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # patching without permissions
        headers = self.create_user_and_login()
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'firstname': 'this-aint-gonna-fly'
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)
        self.assertEqual("Username", user.username)

        # patch as a user of other organization
        rule = Rule.get_by_("user", Scope.ORGANIZATION, Operation.EDIT)
        self.create_user_and_login(rules=[rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'firstname': 'this-aint-gonna-fly'
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)
        self.assertEqual("Username", user.username)

        # patch as another user from the same organization
        rule = Rule.get_by_("user", Scope.OWN, Operation.EDIT)
        self.create_user_and_login(user.organization, [rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'firstname': 'this-aint-gonna-fly'
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)
        self.assertEqual("Username", user.username)

        # edit 'simple' fields
        rule = Rule.get_by_("user", Scope.OWN, Operation.EDIT)
        user.rules.append(rule)
        user.save()
        headers = self.login(user.username)
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'firstname': 'yeah'
        })
        db.session.refresh(user)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual("yeah", user.firstname)

        # edit other user within your organization
        rule = Rule.get_by_("user", Scope.ORGANIZATION, Operation.EDIT)
        headers = self.create_user_and_login(organization=user.organization,
                                             rules=[rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'firstname': 'whatever'
        })
        db.session.refresh(user)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual("whatever", user.firstname)

        # check that password cannot be edited
        rule = Rule.get_by_("user", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'password': 'keep-it-safe'
        })
        self.assertEqual(result.status_code, HTTPStatus.BAD_REQUEST)

        # edit user from different organization, and test other edit fields
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'firstname': 'again',
            'lastname': 'and again',
        })
        db.session.refresh(user)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual("again", user.firstname)
        self.assertEqual("and again", user.lastname)

        # test that you cannot assign rules that you not own
        not_owning_rule = Rule.get_by_("user", Scope.OWN,
                                       Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'rules': [not_owning_rule.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test that you cannot assign role that has rules that you do not own
        role = Role(name="somename", rules=[not_owning_rule],
                    organization=org)
        role.save()
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'rules': [role.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test that you cannot assign rules if you don't have all the rules
        # that the other user has
        headers = self.create_user_and_login(rules=[rule, not_owning_rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'rules': [not_owning_rule.id, rule.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test that you CAN change the rules. To do so, a user is generated
        # that has same rules as current user, but also rule to edit other
        # users and another one current user does not possess
        assigning_user_rules = user.rules
        assigning_user_rules.append(
            Rule.get_by_("user", Scope.GLOBAL, Operation.EDIT)
        )
        assigning_user_rules.append(not_owning_rule)
        headers = self.create_user_and_login(rules=assigning_user_rules)
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'rules': [not_owning_rule.id, rule.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.OK)
        user_rule_ids = [rule['id'] for rule in result.json['rules']]
        self.assertIn(not_owning_rule.id, user_rule_ids)

        # test that you cannot assign roles if you don't have all the
        # permissions for that role yourself (even though you have permission
        # to assign roles)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'roles': [role.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test that you CAN assign roles
        headers = self.create_user_and_login(rules=[rule, not_owning_rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'roles': [role.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.OK)
        user_role_ids = [role['id'] for role in result.json['roles']]
        self.assertIn(role.id, user_role_ids)

        # test that you CANNOT assign roles from different organization
        other_org_role = Role(name="somename", rules=[not_owning_rule],
                              organization=Organization())
        headers = self.create_user_and_login(rules=[rule, not_owning_rule])
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'roles': [other_org_role.id]
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test missing role
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'roles': [-1]
        })
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # test missing rule
        result = self.app.patch(f'/api/user/{user.id}', headers=headers, json={
            'rules': [-1]
        })
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        user.delete()
        role.delete()

    def test_delete_user_permissions(self):

        user = User(firstname="Firstname", lastname="Lastname",
                    username="Username", password="Password", email="a@b.c",
                    organization=Organization())
        user.save()
        self.credentials[user.username] = {'username': user.username,
                                           'password': "Password"}

        # check non-exsitsing user
        headers = self.create_user_and_login()
        result = self.app.delete('/api/user/-1', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # try to delete without any permissions
        headers = self.create_user_and_login()
        result = self.app.delete(f'/api/user/{user.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # same organization but missing permissions
        rule = Rule.get_by_("user", Scope.OWN, Operation.DELETE)
        headers = self.create_user_and_login(organization=user.organization,
                                             rules=[rule])
        result = self.app.delete(f'/api/user/{user.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # other organization with organization scope
        rule = Rule.get_by_("user", Scope.ORGANIZATION,
                            Operation.DELETE)
        headers = self.create_user_and_login(organization=Organization(),
                                             rules=[rule])
        result = self.app.delete(f'/api/user/{user.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # delete yourself
        rule = Rule.get_by_("user", Scope.OWN, Operation.DELETE)
        user.rules.append(rule)
        user.save()
        headers = self.login(user.username)
        result = self.app.delete(f'/api/user/{user.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        # User is deleted by the endpoint! user.delete()

        # delete colleague
        user = User(firstname="Firstname", lastname="Lastname",
                    username="Username", password="Password", email="a@b.c",
                    organization=Organization())
        user.save()
        rule = Rule.get_by_("user", Scope.ORGANIZATION,
                            Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule],
                                             organization=user.organization)
        result = self.app.delete(f'/api/user/{user.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        # User is deleted by the endpoint user.delete()

        # delete as root
        user = User(firstname="Firstname", lastname="Lastname",
                    username="Username", password="Password", email="a@b.c",
                    organization=Organization())
        user.save()
        rule = Rule.get_by_("user", Scope.GLOBAL, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.delete(f'/api/user/{user.id}', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        # user is deleted by endpoint! user.delete()

    def test_view_organization_as_user_permissions(self):

        # view without any permissions
        headers = self.create_user_and_login()
        result = self.app.get('/api/organization', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # view your own organization
        rule = Rule.get_by_("organization", Scope.ORGANIZATION,
                            Operation.VIEW)
        user = self.create_user(rules=[rule])
        headers = self.login(user.username)
        result = self.app.get(f'/api/organization/{user.organization.id}',
                              headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        # try to view another organization without permission
        org = Organization()
        org.save()
        result = self.app.get(f'/api/organization/{org.id}',
                              headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # Missing organization with global view
        rule = Rule.get_by_("organization", Scope.GLOBAL,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.get('/api/organization/-1',
                              headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # test global view
        result = self.app.get(f'/api/organization/{org.id}',
                              headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

    def test_view_organization_as_node_permission(self):
        node, api_key = self.create_node()
        headers = self.login_node(api_key)

        # test list organization with only your organization
        result = self.app.get('/api/organization', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(result.json[0]['id'], node.organization.id)

        # test list organization
        result = self.app.get(
            f'/api/organization/{node.organization.id}',
            headers=headers
        )
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(result.json['id'], node.organization.id)
        node.delete()

    def test_view_organization_as_container_permission(self):
        node, api_key = self.create_node()
        headers = self.login_container(node=node, api_key=api_key)

        # try to get organization where he runs
        result = self.app.get(
            f'/api/organization/{node.organization.id}',
            headers=headers
        )
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertEqual(result.json['id'], node.organization.id)

        # get all organizations in the collaboration
        result = self.app.get(
            '/api/organization',
            headers=headers
        )
        self.assertEqual(result.status_code, HTTPStatus.OK)
        self.assertIsInstance(result.json, list)

        # cleanup
        node.delete()

    def test_create_organization_permissions(self):

        # try creating an organization without permissions
        headers = self.create_user_and_login()
        result = self.app.post('/api/organization', headers=headers, json={
            'name': 'this-aint-gonna-happen'
        })
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # create an organization
        rule = Rule.get_by_("organization", Scope.GLOBAL,
                            Operation.CREATE)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.post('/api/organization', headers=headers, json={
            'name': 'this-is-gonna-happen'
        })
        self.assertEqual(result.status_code, HTTPStatus.CREATED)
        self.assertIsNotNone(Organization.get_by_name("this-is-gonna-happen"))

    def test_patch_organization_permissions(self):

        # unknown organization
        headers = self.create_user_and_login()
        results = self.app.patch('/api/organization/-1', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # try to change anything without permissions
        org = Organization(name="first-name")
        org.save()
        results = self.app.patch(f'/api/organization/{org.id}',
                                 headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # change as super user
        rule = Rule.get_by_("organization", Scope.GLOBAL,
                            Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.patch(f'/api/organization/{org.id}',
                                 headers=headers, json={
                                     "name": "second-name"
                                 })
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['name'], "second-name")

        # change as organization editor
        rule = Rule.get_by_("organization", Scope.ORGANIZATION,
                            Operation.EDIT)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.patch(f'/api/organization/{org.id}',
                                 headers=headers, json={
                                     "name": "third-name"
                                 })
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['name'], "third-name")

        # change other organization as organization editor
        rule = Rule.get_by_("organization", Scope.ORGANIZATION,
                            Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.patch(f'/api/organization/{org.id}',
                                 headers=headers, json={
                                     "name": "third-name"
                                 })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

    def test_organization_view_nodes(self):

        # create organization, collaboration and node
        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()
        node = Node(organization=org, collaboration=col)
        node.save()

        # try to view without permissions
        headers = self.create_user_and_login(org)
        results = self.app.get(f"/api/organization/{org.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to view with organization permissions
        rule = Rule.get_by_("node", Scope.ORGANIZATION, Operation.VIEW)
        headers = self.create_user_and_login(org, rules=[rule])
        results = self.app.get(f"/api/organization/{org.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # try to view other organization
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/organization/{org.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to view with global permissions
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/organization/{org.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # try to view as node
        headers = self.create_node_and_login(organization=org)
        results = self.app.get(f"/api/organization/{org.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # try to view as node from another organization
        headers = self.create_node_and_login()
        results = self.app.get(f"/api/organization/{org.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # cleanup
        node.delete()

    def test_organization_view_collaboration_permissions(self):

        # test unknown organization
        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        headers = self.create_user_and_login()
        results = self.app.get('/api/organization/-1/collaboration',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # test view without any permission
        results = self.app.get(f'/api/organization/{org.id}/collaboration',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test view with organization scope
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.get(f'/api/organization/{org.id}/collaboration',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test view with organization scope other organiation
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f'/api/organization/{org.id}/collaboration',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test view with global scope
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f'/api/organization/{org.id}/collaboration',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test as node
        headers = self.create_node_and_login(organization=org,
                                             collaboration=col)
        results = self.app.get(f'/api/organization/{org.id}/collaboration',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_view_collaboration_permissions(self):

        # setup organization and collaboration
        org = Organization()
        col = Collaboration(organizations=[org])
        col.save()

        # try view the collaboration without any permissions
        headers = self.create_user_and_login(organization=org)
        results = self.app.get(f"/api/collaboration/{col.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to view it with organization permissions
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # try to view it from an outside organization
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # view it with global view permissions
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test access as node
        headers = self.create_node_and_login(organization=org,
                                             collaboration=col)
        results = self.app.get(f"/api/collaboration/{col.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test access as container
        headers = self.login_container(collaboration=col, organization=org)
        results = self.app.get(f"/api/collaboration/{col.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        org.delete()
        col.delete()

    def test_edit_collaboration_permissions(self):

        # test an unknown collaboration
        headers = self.create_user_and_login()
        results = self.app.patch("/api/collaboration/-1", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # test editing without any permission
        col = Collaboration(name="collaboration-1")
        col.save()
        headers = self.create_user_and_login()
        results = self.app.patch(f"/api/collaboration/{col.id}",
                                 headers=headers, json={
                                     "name": "this-aint-gonna-fly"
                                 })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test editing with global permissions
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.patch(f"/api/collaboration/{col.id}",
                                 headers=headers, json={
                                     "name": "this-is-gonna-fly"
                                 })
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json["name"], "this-is-gonna-fly")

    def test_delete_collaboration_permissions(self):

        col = Collaboration()
        col.save()

        # test deleting unknown collaboration
        headers = self.create_user_and_login()
        results = self.app.delete("/api/collaboration/-1",
                                  headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # test deleting without permission
        results = self.app.delete(f"/api/collaboration/{col.id}",
                                  headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test deleting with permission
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f"/api/collaboration/{col.id}",
                                  headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_view_collaboration_organization_permissions_as_user(self):
        headers = self.create_user_and_login()

        # un-existing collaboration
        results = self.app.get("/api/collaboration/-1/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        # access without the proper permissions
        headers = self.create_user_and_login(organization=org)
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # global permissions
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # organization permissions of another organization
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # now with the correct organization
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_view_collaboration_organization_permissions_as_node(self):

        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        # node of a different organization
        headers = self.create_node_and_login()
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # node of the correct organization
        headers = self.create_node_and_login(organization=org,
                                             collaboration=col)
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_view_collaboration_organization_permissions_as_container(self):

        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        # node of a different organization
        headers = self.login_container()
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        headers = self.login_container(organization=org, collaboration=col)
        results = self.app.get(f"/api/collaboration/{col.id}/organization",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_edit_collaboration_organization_permissions(self):

        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        org2 = Organization()
        org2.save()

        # try to do it without permission
        headers = self.create_user_and_login()
        results = self.app.post(f"/api/collaboration/{col.id}/organization",
                                headers=headers, json={'id': org2.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # edit permissions
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.post(f"/api/collaboration/{col.id}/organization",
                                headers=headers, json={'id': org2.id})
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), 2)

    def test_delete_collaboration_organization_pesmissions(self):

        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        # try to do it without permission
        headers = self.create_user_and_login()
        results = self.app.delete(f"/api/collaboration/{col.id}/organization",
                                  headers=headers, json={'id': org.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # delete it!
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f"/api/collaboration/{col.id}/organization",
                                  headers=headers, json={'id': org.id})
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json, [])

    def test_view_collaboration_node_permissions(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        node = Node(collaboration=col, organization=org)
        node.save()

        # try to view an non-existant collaboration
        headers = self.create_user_and_login()
        results = self.app.get("/api/collaboration/-1/node", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # try to view without any permissions
        results = self.app.get(f"/api/collaboration/{col.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to view from another organzization
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to view from another organization with global permissions
        rule = Rule.get_by_("collaboration", Scope.GLOBAL,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f"/api/collaboration/{col.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), len(col.nodes))

        # try to view from your organization
        rule = Rule.get_by_("collaboration", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule], organization=org)
        results = self.app.get(f"/api/collaboration/{col.id}/node",
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # cleanup
        node.delete()

    def test_add_collaboration_node_permissions(self):

        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()
        node = Node(organization=org)
        node.save()

        # try non-existant collaboration
        headers = self.create_user_and_login()

        results = self.app.post('/api/collaboration/-1/node', headers=headers,
                                json={'id': node.id})

        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # try without proper permissions
        results = self.app.post(f'/api/collaboration/{col.id}/node',
                                headers=headers, json={'id': node.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to add non-existing node
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.post(f'/api/collaboration/{col.id}/node',
                                headers=headers, json={'id': -1})
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # add a node!
        results = self.app.post(f'/api/collaboration/{col.id}/node',
                                headers=headers, json={'id': node.id})
        self.assertEqual(results.status_code, HTTPStatus.CREATED)
        self.assertEqual(len(results.json), len(col.nodes))

        # try to add a node thats already in there
        results = self.app.post(f'/api/collaboration/{col.id}/node',
                                headers=headers, json={'id': node.id})
        self.assertEqual(results.status_code, HTTPStatus.BAD_REQUEST)

        # cleanup
        node.delete()

    def test_delete_collaboration_node_permissions(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        node = Node(organization=org, collaboration=col)
        node.save()

        # try non-existant collaboration
        headers = self.create_user_and_login()
        results = self.app.delete('/api/collaboration/-1/node',
                                  headers=headers, json={'id': node.id})
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # try without proper permissions
        results = self.app.delete(f'/api/collaboration/{col.id}/node',
                                  headers=headers, json={'id': node.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # try to add non-existing node
        rule = Rule.get_by_("collaboration", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f'/api/collaboration/{col.id}/node',
                                  headers=headers, json={'id': -1})
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # try to add a node thats not in there
        node2 = Node()
        node2.save()
        results = self.app.delete(f'/api/collaboration/{col.id}/node',
                                  headers=headers, json={'id': node2.id})
        self.assertEqual(results.status_code, HTTPStatus.BAD_REQUEST)

        # delete a node!
        results = self.app.delete(f'/api/collaboration/{col.id}/node',
                                  headers=headers, json={'id': node.id})
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # cleanup
        node.delete()
        node2.delete()

    def test_view_collaboration_task_permissions_as_user(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        task = Task(collaboration=col)
        task.save()

        # view non existing collaboration
        headers = self.create_user_and_login()
        results = self.app.get('/api/collaboration/-1/task',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # view without any permission
        results = self.app.get(f'/api/collaboration/{col.id}/task',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # view from another organization
        rule = Rule.get_by_("task", Scope.ORGANIZATION,
                            Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f'/api/collaboration/{col.id}/task',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # view from your own organization
        headers = self.create_user_and_login(rules=[rule], organization=org)
        results = self.app.get(f'/api/collaboration/{col.id}/task',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), len(col.tasks))

        # view with global permissions
        rule = Rule.get_by_("task", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f'/api/collaboration/{col.id}/task',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), len(col.tasks))

    def test_view_collaboration_task_permissions_as_node(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        node, api_key = self.create_node(org, col)
        task = Task(collaboration=col)
        task.save()

        headers = self.login_node(api_key)
        results = self.app.get(f'/api/collaboration/{col.id}/task',
                               headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # cleanup
        node.delete()

    def test_view_node_permissions_as_user(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        node = Node(organization=org, collaboration=col)
        node.save()

        # view non existing node
        headers = self.create_user_and_login()
        results = self.app.get('/api/node/-1', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # missing permissions
        results = self.app.get(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # organization permissions
        rule1 = Rule.get_by_("node", Scope.ORGANIZATION, Operation.VIEW)
        headers = self.create_user_and_login(organization=org, rules=[rule1])
        results = self.app.get(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # organization permissions from another organization
        headers = self.create_user_and_login(rules=[rule1])
        results = self.app.get(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # global permissions
        rule2 = Rule.get_by_("node", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule2])
        results = self.app.get(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # list organization permissions
        headers = self.create_user_and_login(organization=org, rules=[rule1])
        results = self.app.get('/api/node', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), len(col.nodes))

        # list global permissions
        headers = self.create_user_and_login(rules=[rule2])
        results = self.app.get('/api/node', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), len(Node.get()))

        # cleanup
        node.delete()

    def test_view_node_permissions_as_node(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        node, api_key = self.create_node(org, col)

        headers = self.login_node(api_key)

        # global permissions
        results = self.app.get(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # list organization permissions
        results = self.app.get('/api/node', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(len(results.json), len(org.nodes))

        # cleanup
        node.delete()

    def test_create_node_permissions(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        org2 = Organization()
        org2.save()

        # test non existing collaboration
        headers = self.create_user_and_login()
        results = self.app.post('/api/node', headers=headers,
                                json={'collaboration_id': -1})
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # test creating a node without any permissions
        headers = self.create_user_and_login()
        results = self.app.post('/api/node', headers=headers,
                                json={'collaboration_id': col.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # testing creating a node with organization permissions and supplying
        # an organization id
        rule = Rule.get_by_("node", Scope.ORGANIZATION, Operation.CREATE)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.post('/api/node', headers=headers, json={
            'collaboration_id': col.id,
            'organization_id': org.id
        })

        self.assertEqual(results.status_code, HTTPStatus.CREATED)
        node_id = results.json.get('id')
        results = self.app.post('/api/node', headers=headers, json={
            'collaboration_id': col.id,
            'organization_id': org2.id  # <-------
        })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test adding a node to an collaboration from an organization witch
        # does not belong to the collaboration
        headers = self.create_user_and_login(organization=org2, rules=[rule])
        results = self.app.post('/api/node', headers=headers, json={
            'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.BAD_REQUEST)

        # check an creating an already existing node
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.post('/api/node', headers=headers, json={
            'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.BAD_REQUEST)

        # lets retry that
        node = Node.get(node_id)
        node.delete()
        results = self.app.post('/api/node', headers=headers, json={
            'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.CREATED)

        # test global permissions
        col.organizations.append(org2)
        col.save()
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.CREATE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.post('/api/node', headers=headers, json={
            'collaboration_id': col.id,
            'organization_id': org2.id
        })
        self.assertEqual(results.status_code, HTTPStatus.CREATED)

    def test_delete_node_permissions(self):

        org = Organization()
        col = Collaboration(organizations=[org])
        node = Node(organization=org, collaboration=col)
        node.save()

        # unexisting node
        headers = self.create_user_and_login()
        results = self.app.delete('/api/node/-1', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # organization permission other organization
        rule = Rule.get_by_('node', Scope.ORGANIZATION, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # organization permission
        rule = Rule.get_by_('node', Scope.ORGANIZATION, Operation.DELETE)
        headers = self.create_user_and_login(organization=org, rules=[rule])
        results = self.app.delete(f'/api/node/{node.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # global permission
        node2 = Node(organization=org, collaboration=col)
        node2.save()
        rule = Rule.get_by_('node', Scope.GLOBAL, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f'/api/node/{node2.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_patch_node_permissions_as_user(self):
        # test patching non-existant node
        headers = self.create_user_and_login()
        results = self.app.patch("/api/node/-1", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # test user without any permissions
        org = Organization()
        col = Collaboration(organizations=[org])
        node = Node(organization=org, collaboration=col)
        node.save()

        results = self.app.patch(f"/api/node/{node.id}", headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test user with global permissions
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.patch(f"/api/node/{node.id}", headers=headers,
                                 json={"name": "A"})
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['name'], "A")

        # test user with org permissions and own organization
        rule = Rule.get_by_("node", Scope.ORGANIZATION, Operation.EDIT)
        headers = self.create_user_and_login(org, [rule])
        results = self.app.patch(f"/api/node/{node.id}", headers=headers,
                                 json={'name': 'B'})
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['name'], "B")

        # test user with org permissions and other organization
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.patch(f"/api/node/{node.id}", headers=headers,
                                 json={'name': 'C'})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test updatin the `organization_id` with organization permissions
        org2 = Organization()
        org2.save()
        results = self.app.patch(f"/api/node/{node.id}", headers=headers,
                                 json={'organization_id': org2.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test assigning it to a node thats not part of the collaborat
        col2 = Collaboration(organizations=[org2])
        col2.save()
        results = self.app.patch(f"/api/node/{node.id}", headers=headers,
                                 json={'collaboration_id': col2.id})
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # collaboration_id and organization_id. Note that the organization
        # is assigned before the collaboration is defined.
        rule = Rule.get_by_("node", Scope.GLOBAL, Operation.EDIT)
        headers = self.create_user_and_login(org2, rules=[rule])
        results = self.app.patch(f'/api/node/{node.id}', headers=headers,
                                 json={'collaboration_id': col2.id,
                                       'organization_id': org2.id})
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['organization']['id'], org2.id)
        self.assertEqual(results.json['collaboration']['id'], col2.id)

        # try to patch the node's VPN IP address
        results = self.app.patch(f'/api/node/{node.id}', headers=headers,
                                 json={'ip':'0.0.0.0'})
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['ip'], '0.0.0.0')

        # assign unknow organization
        results = self.app.patch(f'/api/node/{node.id}', headers=headers,
                                 json={'organization_id': -1})
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # cleanup
        node.delete()

    def test_view_task_permissions_as_user(self):
        # non existing task
        headers = self.create_user_and_login()
        results = self.app.get('/api/task/-1', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # test user without any permissions and id
        org = Organization()
        col = Collaboration(organizations=[org])
        task = Task(name="unit", collaboration=col)
        task.save()

        results = self.app.get(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test user with org permissions with id
        rule = Rule.get_by_("task", Scope.ORGANIZATION, Operation.VIEW)
        headers = self.create_user_and_login(org, rules=[rule])
        results = self.app.get(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertEqual(results.json['name'], 'unit')

        # test user with org permissions with id from another org
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test user with org permissions without id
        headers = self.create_user_and_login(org, rules=[rule])
        results = self.app.get('/api/task', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test user with global permissions and id
        rule = Rule.get_by_("task", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.get(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test user with global permissions without id
        results = self.app.get('/api/task', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_view_task_permissions_as_node_and_container(self):
        # test node with id
        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()
        task = Task(collaboration=col, image="some-image")
        task.save()
        res = Result(task=task)
        res.save()

        headers = self.create_node_and_login(organization=org)
        results = self.app.get(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test node without id
        results = self.app.get('/api/task', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test container with id
        headers = self.login_container(collaboration=col, organization=org,
                                       task=task)
        results = self.app.get(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test container without id
        results = self.app.get('/api/task', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

    def test_create_task_permission_as_user(self):
        # non existant collaboration
        headers = self.create_user_and_login()
        results = self.app.post('/api/task', headers=headers, json={
            "collaboration_id": -1
        })
        self.assertEqual(results.status_code, HTTPStatus.NOT_FOUND)

        # organizations outside of collaboration
        org = Organization()
        org.save()
        col = Collaboration(organizations=[org])
        col.save()

        # task without any node created
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            "collaboration_id": col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.BAD_REQUEST)

        # node is used implicitly as in further checks, can only create task
        # if node has been created
        node = Node(organization=org, collaboration=col)

        org2 = Organization()
        org2.save()

        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org2.id}], 'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.BAD_REQUEST)

        # user without any permissions
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            "collaboration_id": col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # user with organization permissions for other organization
        rule = Rule.get_by_("task", Scope.ORGANIZATION, Operation.CREATE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}], 'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # user with organization permissions
        headers = self.create_user_and_login(org, rules=[rule])
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}], 'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.CREATED)

        # user with global permissions but outside of the collaboration. They
        # should *not* be allowed to create a task in a collaboration that
        # they're not a part of
        # TODO add test for user with global permission that creates a task for
        # another organization than their own in the same collaboration
        rule = Rule.get_by_("task", Scope.GLOBAL, Operation.CREATE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}], 'collaboration_id': col.id
        })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test master task
        rule = Rule.get_by_("task", Scope.ORGANIZATION, Operation.CREATE)
        headers = self.create_user_and_login(org, rules=[rule])
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            'collaboration_id': col.id,
            'master': True
        })
        self.assertEqual(results.status_code, HTTPStatus.CREATED)

        # cleanup
        node.delete()

    def test_create_task_permissions_as_container(self):
        org = Organization()
        col = Collaboration(organizations=[org])
        parent_task = Task(collaboration=col, image="some-image")
        parent_task.save()
        parent_res = Result(organization=org, task=parent_task)
        parent_res.save()

        # test wrong image name
        headers = self.login_container(collaboration=col, organization=org,
                                       task=parent_task)
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            'collaboration_id': col.id,
            'image': 'other-image'
        })

        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test other collaboration_id
        col2 = Collaboration(organizations=[org])
        col2.save()
        node2 = Node(organization=org, collaboration=col2)
        node2.save()
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            'collaboration_id': col2.id,
            'image': 'some-image'
        })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test with correct parameters
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            'collaboration_id': col.id,
            'image': 'some-image'
        })
        self.assertEqual(results.status_code, HTTPStatus.CREATED)

        # test already completed task
        parent_res.finished_at = datetime.date(2020, 1, 1)
        parent_res.save()
        results = self.app.post('/api/task', headers=headers, json={
            "organizations": [{'id': org.id}],
            'collaboration_id': col.id,
            'image': 'some-image'
        })
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # cleanup
        node2.delete()

    def test_delete_task_permissions(self):

        # test non-existing task
        headers = self.create_user_and_login()
        self.app.delete('/api/task/-1', headers=headers)

        # test with organization permissions from other organization
        org = Organization()
        col = Collaboration(organizations=[org])
        task = Task(collaboration=col)
        task.save()

        rule = Rule.get_by_("task", Scope.ORGANIZATION, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.UNAUTHORIZED)

        # test with organization permissions
        headers = self.create_user_and_login(org, [rule])
        results = self.app.delete(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test with global permissions
        task = Task(collaboration=col)
        task.save()
        rule = Rule.get_by_("task", Scope.GLOBAL, Operation.DELETE)
        headers = self.create_user_and_login(rules=[rule])
        results = self.app.delete(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)

        # test that all results are also deleted
        task = Task(collaboration=col)
        res = Result(task=task)
        res.save()
        result_id = res.id  # cannot access this after deletion
        results = self.app.delete(f'/api/task/{task.id}', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
        self.assertIsNone(Task.get(result_id))

    def test_view_task_result_permissions_as_user(self):

        # non-existing task
        headers = self.create_user_and_login()
        result = self.app.get('/api/task/-1/result', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.NOT_FOUND)

        # test with organization permissions from other organization
        org = Organization()
        col = Collaboration(organizations=[org])
        task = Task(collaboration=col)
        # NB: node is used implicitly in task/{id}/result schema
        node = Node(organization=org, collaboration=col)
        res = Result(task=task, organization=org)
        res.save()

        rule = Rule.get_by_("result", Scope.ORGANIZATION, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.get(f'/api/task/{task.id}/result', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.UNAUTHORIZED)

        # test with organization permission
        headers = self.create_user_and_login(org, [rule])
        result = self.app.get(f'/api/task/{task.id}/result', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        # test with global permission
        rule = Rule.get_by_("result", Scope.GLOBAL, Operation.VIEW)
        headers = self.create_user_and_login(rules=[rule])
        result = self.app.get(f'/api/task/{task.id}/result', headers=headers)
        self.assertEqual(result.status_code, HTTPStatus.OK)

        # cleanup
        node.delete()

    def test_view_task_result_permissions_as_container(self):
        # test if container can
        org = Organization()
        col = Collaboration(organizations=[org])
        task = Task(collaboration=col, image="some-image")
        task.save()
        res = Result(task=task, organization=org)
        res.save()

        headers = self.login_container(collaboration=col, organization=org,
                                       task=task)
        results = self.app.get(f'/api/task/{task.id}/result', headers=headers)
        self.assertEqual(results.status_code, HTTPStatus.OK)
